import os
from pathlib import Path
from shutil import disk_usage
from typing import Any

from ultima_scraper_api import user_types
from ultima_scraper_db.databases.ultima_archive.schemas.templates.site import (
    UserModel as DBUserModel,
)

from ultima_scraper_collection.config import Directory


def check_space(
    custom_directory: list[Directory],
):
    root = ""
    while not root:
        paths: list[dict[str, Any]] = []
        for directory in custom_directory:
            # ISSUE
            # Could cause problems w/ relative/symbolic links that point to another hard drive
            # Haven't tested if it calculates hard A or relative/symbolic B's total space.
            # size is in GB
            assert directory.path
            obj_Disk = disk_usage(str(directory.path.parent))
            free = obj_Disk.free / (1024.0**3)
            x = {}
            x["path"] = directory.path
            x["free"] = free
            x["min_space"] = directory.minimum_space
            paths.append(x)
        for item in paths:
            download_path = item["path"]
            free = item["free"]
            if free > item["min_space"]:
                root = download_path
                break
    return root


from ultima_scraper_api.apis.onlyfans.classes.user_model import (
    create_user as OFUserModel,
)


async def is_valuable(user: DBUserModel | user_types):
    # Checks if performer has active subscription or has supplied content to a buyer
    if isinstance(user, DBUserModel):
        if await user.find_buyers(active=True):
            return True
        else:
            return False
    else:
        if user.is_performer():
            if isinstance(user, OFUserModel):
                if (
                    user.subscribed_is_expired_now == False
                    or await user.get_paid_contents()
                ):
                    return True
                else:
                    return False
            else:
                # We need to add paid_content checker
                if user.following:
                    return True
                else:
                    return False
        else:
            return False


async def is_notif_valuable(api_user: user_types):
    if await is_valuable(api_user):
        if await api_user.subscription_price() == 0:
            if isinstance(api_user, OFUserModel) and await api_user.get_paid_contents():
                return True
            return False
        else:
            return True
    return False


async def walk(directory: Path):
    all_files: list[Path] = []
    for root, _subdirs, files in os.walk(directory):
        x = [Path(root, x) for x in files]
        all_files.extend(x)
    return all_files


def find_unused_filename(filepath: Path):
    base_name = filepath.stem  # Get the filename without extension
    extension = filepath.suffix  # Get the file extension
    counter = 2

    while filepath.exists():
        new_name = f"{base_name} ({counter}){extension}"
        filepath = filepath.with_name(new_name)
        counter += 1

    return filepath



from pathlib import Path
from typing import TYPE_CHECKING, Any

from ultima_scraper_api.apis.fansly.fansly import FanslyAPI
from ultima_scraper_renamer.reformat import ReformatManager

from ultima_scraper_collection.config import Sites
from ultima_scraper_collection.managers.metadata_manager.metadata_manager import (
    ApiExtractor,
    ContentMetadata,
)
from ultima_scraper_collection.managers.option_manager import OptionManager
from ultima_scraper_collection.managers.server_manager import ServerManager
from ultima_scraper_collection.modules.module_streamliner import StreamlinedDatascraper

if TYPE_CHECKING:
    from ultima_scraper_api.apis.fansly.classes.auth_model import AuthModel
    from ultima_scraper_api.apis.fansly.classes.message_model import create_message
    from ultima_scraper_api.apis.fansly.classes.post_model import create_post
    from ultima_scraper_api.apis.fansly.classes.story_model import create_story
    from ultima_scraper_api.apis.fansly.classes.user_model import create_user


class FanslyDataScraper(StreamlinedDatascraper):
    def __init__(
        self,
        api: FanslyAPI,
        option_manager: OptionManager,
        server_manager: ServerManager,
        site_config: Sites.FanslyAPIConfig,
    ) -> None:
        self.api = api
        self.option_manager = option_manager
        self.site_config = site_config
        StreamlinedDatascraper.__init__(self, self, server_manager)

    # Scrapes the API for content
    async def media_scraper(
        self,
        content_result: "create_story | create_post | create_message",
        subscription: "create_user",
        api_type: str,
    ) -> dict[str, Any]:
        authed = subscription.get_authed()
        site_config = self.site_config
        new_set: dict[str, Any] = {"content": []}
        directories: list[Path] = []
        if api_type == "Stories":
            pass
        if api_type == "Archived":
            pass
        if api_type == "Posts":
            pass
        if api_type == "Messages":
            pass

        content_metadata = ContentMetadata(content_result.id, api_type)
        await content_metadata.resolve_extractor(ApiExtractor(content_result))
        for asset in content_metadata.medias:
            if asset.urls:
                reformat_manager = ReformatManager(authed, self.filesystem_manager)
                reformat_item = reformat_manager.prepare_reformat(asset)
                file_directory = reformat_item.reformat(
                    site_config.download_setup.directory_format
                )
                reformat_item.directory = file_directory
                file_path = reformat_item.reformat(
                    site_config.download_setup.filename_format
                )
                asset.directory = file_directory
                asset.filename = file_path.name

                if file_directory not in directories:
                    directories.append(file_directory)
        new_set["content"].append(content_metadata)
        new_set["directories"] = directories
        return new_set

    async def get_all_stories(self, subscription: "create_user"):
        """
        get_all_stories(subscription: create_user)

        This function returns a list of all stories and archived stories from the given subscription.

        Arguments:
        subscription (create_user): An instance of the create_user class.

        Returns:
        list[create_story]: A list containing all stories and archived stories from the subscription.
        """
        master_set: list["create_story"] = []
        master_set.extend(await subscription.get_stories())
        # master_set.extend(await subscription.get_archived_stories())
        return master_set

    async def get_all_posts(self, subscription: "create_user"):
        temp_master_set = await subscription.get_posts()
        collections = await subscription.get_collections()
        for collection in collections:
            temp_master_set.append(
                await subscription.get_collection_content(collection)
            )
        return temp_master_set

    async def get_all_subscriptions(
        self,
        authed: "AuthModel",
        identifiers: list[int | str] = [],
        refresh: bool = True,
    ):
        """
        get_all_subscriptions(authed: AuthModel, identifiers: list[int | str] = [], refresh: bool = True)

        This function returns a list of all subscriptions, including both subscriptions and followings,
        from the given authenticated user.

        Arguments:
        authed (AuthModel): An instance of the AuthModel class.
        identifiers (list[int | str], optional): A list of identifiers (username or id) for the subscriptions. Defaults to an empty list.
        refresh (bool, optional): A flag indicating whether to refresh the list of subscriptions. Defaults to True.

        Returns:
        list[create_subscription]: A list of all subscriptions, including both subscriptions and followings, from the authenticated user.
        """
        authed.followed_users = await authed.get_followings(identifiers=identifiers)
        subscriptions = await authed.get_subscriptions(
            identifiers=identifiers, refresh=refresh, sub_type="active"
        )
        subscriptions.sort(key=lambda x: x.ends_at)
        return subscriptions



# type: ignore
"""content

Revision ID: 3076beb33c1b
Revises: 
Create Date: 2021-01-08 23:09:26.868834

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "3076beb33c1b"
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "medias",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("media_id", sa.Integer(), nullable=True),
        sa.Column("post_id", sa.Integer(), nullable=False),
        sa.Column("link", sa.String(), nullable=True),
        sa.Column("directory", sa.String(), nullable=True),
        sa.Column("filename", sa.String(), nullable=True),
        sa.Column("size", sa.Integer(), nullable=True),
        sa.Column("media_type", sa.String(), nullable=True),
        sa.Column("downloaded", sa.Integer(), nullable=True),
        sa.Column("created_at", sa.DATETIME(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("media_id"),
    )
    op.create_table(
        "stories",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("post_id", sa.Integer(), nullable=False),
        sa.Column("text", sa.String(), nullable=True),
        sa.Column("price", sa.Integer(), nullable=True),
        sa.Column("paid", sa.Integer(), nullable=True),
        sa.Column("created_at", sa.DATETIME(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("post_id"),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("stories")
    op.drop_table("medias")
    # ### end Alembic commands ###



from __future__ import annotations

import copy
import os
import shutil
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING, Any, Generator, Literal

import ultima_scraper_api
from aiohttp.client_reqrep import ClientResponse
from ultima_scraper_api.helpers.main_helper import open_partial
from ultima_scraper_api.managers.session_manager import EXCEPTION_TEMPLATE
from ultima_scraper_collection.helpers import main_helper
from ultima_scraper_renamer.reformat import FormatAttributes, ReformatItem

if TYPE_CHECKING:
    api_types = ultima_scraper_api.api_types
    user_types = ultima_scraper_api.user_types
    from ultima_scraper_collection import datascraper_types
    from ultima_scraper_collection.config import site_config_types


class FilesystemManager:
    def __init__(self) -> None:
        self.user_data_directory = Path("__user_data__")
        self.trash_directory = self.user_data_directory.joinpath("trash")
        self.profiles_directory = self.user_data_directory.joinpath("profiles")
        self.devices_directory = self.user_data_directory.joinpath("drm_device")
        self.settings_directory = Path("__user_data__")
        self.ignore_files = ["desktop.ini", ".DS_Store", ".DS_store", "@eaDir"]
        self.directory_manager: DirectoryManager | None = None
        self.directory_manager_users: dict[int, DirectoryManager] = {}
        self.file_manager_users: dict[int, FileManager] = {}

    def __iter__(self):
        for each in self.__dict__.values():
            yield each

    def check(self):
        for directory in self:
            if isinstance(directory, Path):
                directory.mkdir(exist_ok=True)

    def move(self, src: Path, trg: Path):
        shutil.move(src, trg)

    def remove_mandatory_files(
        self, files: list[Path] | Generator[Path, None, None], keep: list[str] = []
    ):
        folders = [x for x in files if x.name not in self.ignore_files]
        if keep:
            folders = [x for x in files if x.name in keep]
        return folders

    def get_directory_manager(self, user_id: int):
        return self.directory_manager_users[user_id]

    def get_file_manager(self, user_id: int):
        return self.file_manager_users[user_id]

    def activate_directory_manager(self, site_config: site_config_types):
        from ultima_scraper_collection.helpers import main_helper

        root_metadata_directory = main_helper.check_space(
            site_config.metadata_setup.directories
        )
        root_download_directory = main_helper.check_space(
            site_config.download_setup.directories
        )
        self.directory_manager = DirectoryManager(
            site_config,
            root_metadata_directory,
            root_download_directory,
        )

    def trash(self):
        pass

    async def write_data(
        self, response: ClientResponse, download_path: Path, callback: Any = None
    ):
        status_code = None
        if response.status == 200:
            total_length = 0
            os.makedirs(os.path.dirname(download_path), exist_ok=True)
            with open_partial(download_path) as f:
                partial_path = f.name
                try:
                    async for data in response.content.iter_chunked(4096):
                        f.write(data)
                        length = len(data)
                        total_length += length
                        if callback:
                            callback(length)
                except EXCEPTION_TEMPLATE as _e:
                    status_code = 1
                except Exception as _e:
                    raise Exception(f"Unknown Error: {_e}")
                except:
                    os.unlink(partial_path)
                    raise
                else:
                    if status_code:
                        os.unlink(partial_path)
                    else:
                        try:
                            os.replace(partial_path, download_path)
                        except OSError:
                            pass
        else:
            if response.content_length:
                pass
                # progress_bar.update_total_size(-response.content_length)
            status_code = 2
        return status_code

    async def create_option(
        self,
        datascraper: datascraper_types,
        username: str,
        directory: Path,
        format_key: str,
    ):
        api = datascraper.api
        option = {
            "site_name": api.site_name,
            "profile_username": username,
            "model_username": username,
            "directory": directory,
        }
        reformat_item_fd = ReformatItem(option)
        assert self.directory_manager
        f_d_p = reformat_item_fd.remove_non_unique(self.directory_manager, format_key)
        return f_d_p

    async def update_performer_folder_name(
        self, datascraper: datascraper_types, user: user_types
    ):
        for directory in datascraper.site_config.download_setup.directories:
            assert directory.path
            for username in user.aliases:
                f_d_p = await self.create_option(
                    datascraper, username, directory.path, "file_directory_format"
                )
                if f_d_p.exists():
                    new_folderpath = await self.create_option(
                        datascraper,
                        user.username,
                        directory.path,
                        "file_directory_format",
                    )
                    f_d_p.rename(new_folderpath)
                    # can merge duplicate folders here
                    break

    async def create_directory_manager(
        self, site_config: site_config_types, user: user_types
    ):
        api = user.get_api()
        if self.directory_manager:
            final_download_directory = self.directory_manager.root_download_directory
            for directory in site_config.download_setup.directories:
                option = {
                    "site_name": api.site_name,
                    "profile_username": user.username,
                    "model_username": user.username,
                    "directory": directory.path,
                }
                reformat_item_fd = ReformatItem(option)
                f_d_p = reformat_item_fd.remove_non_unique(
                    self.directory_manager, "file_directory_format"
                )
                if f_d_p.exists():
                    final_download_directory = directory.path
                    assert final_download_directory
                    break
            directory_manager = DirectoryManager(
                site_config,
                self.directory_manager.root_metadata_directory,
                final_download_directory,
            )
            # option = {
            #     "site_name": api.site_name,
            #     "profile_username": user.username,
            #     "model_username": user.username,
            #     "directory": directory_manager.root_download_directory,
            # }
            # reformat_item_fd = ReformatItem(option)
            # _f_d_p = reformat_item_fd.remove_non_unique(
            #     directory_manager, "file_directory_format"
            # )
            # # f_d_p.mkdir(parents=True, exist_ok=True)
            # option["directory"] = self.directory_manager.root_metadata_directory
            # reformat_item_md = ReformatItem(option)
            # _f_d_p_2 = reformat_item_md.remove_non_unique(
            #     self.directory_manager, "metadata_directory_format"
            # )
            # # f_d_p_2.mkdir(parents=True, exist_ok=True)
            self.directory_manager_users[user.id] = directory_manager
            self.file_manager_users[user.id] = FileManager(directory_manager)
            return directory_manager

    async def format_directories(self, subscription: user_types) -> DirectoryManager:
        directory_manager = self.get_directory_manager(subscription.id)
        site_config = directory_manager.site_config
        file_manager = self.get_file_manager(subscription.id)
        authed = subscription.get_authed()
        api = authed.api
        authed_username = authed.username
        subscription_username = subscription.username
        site_name = authed.api.site_name
        reformat_item = ReformatItem()
        reformat_item.site_name = site_name
        reformat_item.profile_username = authed_username
        reformat_item.model_username = subscription_username
        reformat_item.date = datetime.today()
        download_setup = site_config.download_setup
        reformat_item.date_format = download_setup.date_format
        reformat_item.text_length = download_setup.text_length
        reformat_item.directory = directory_manager.root_metadata_directory
        metadata_setup = site_config.metadata_setup
        string = reformat_item.reformat(metadata_setup.directory_format)
        directory_manager.user.metadata_directory = Path(string)
        reformat_item_2 = copy.copy(reformat_item)
        reformat_item_2.directory = directory_manager.root_download_directory
        string = reformat_item_2.reformat(metadata_setup.directory_format)
        formtatted_root_download_directory = reformat_item_2.remove_non_unique(
            directory_manager, "file_directory_format"
        )
        directory_manager.user.download_directory = formtatted_root_download_directory
        await file_manager.set_default_files(reformat_item, reformat_item_2)
        _metadata_filepaths = await file_manager.find_metadata_files(legacy_files=False)
        # I forgot why we need to set default file twice
        await file_manager.set_default_files(reformat_item, reformat_item_2)
        user_metadata_directory = directory_manager.user.metadata_directory
        _user_download_directory = directory_manager.user.download_directory
        legacy_metadata_directory = user_metadata_directory
        directory_manager.user.legacy_metadata_directories.append(
            legacy_metadata_directory
        )
        items = api.CategorizedContent()
        for api_type, _ in items:
            legacy_metadata_directory_2 = user_metadata_directory.joinpath(api_type)
            directory_manager.user.legacy_metadata_directories.append(
                legacy_metadata_directory_2
            )
        legacy_model_directory = directory_manager.root_download_directory.joinpath(
            site_name, subscription_username
        )
        directory_manager.user.legacy_download_directories.append(
            legacy_model_directory
        )
        return directory_manager


class DirectoryManager:
    def __init__(
        self,
        site_config: site_config_types,
        root_metadata_directory: Path = Path(),
        root_download_directory: Path = Path(),
    ) -> None:
        self.root_directory = Path()
        self.root_metadata_directory = Path(root_metadata_directory)
        self.root_download_directory = Path(root_download_directory)
        self.user = self.UserDirectories()
        self.site_config = site_config
        formats = FormatTypes(site_config)
        string, status = formats.check_rules()
        if not status:
            print(string)
            exit(0)
        self.formats = formats
        pass

    def create_directories(self):
        # self.profile.create_directories()
        self.root_metadata_directory.mkdir(exist_ok=True)
        self.root_download_directory.mkdir(exist_ok=True)

    def delete_empty_directories(
        self, directory: Path, filesystem_manager: FilesystemManager
    ):
        for root, dirnames, _files in os.walk(directory, topdown=False):
            for dirname in dirnames:
                full_path = os.path.realpath(os.path.join(root, dirname))
                contents = os.listdir(full_path)
                if not contents:
                    shutil.rmtree(full_path, ignore_errors=True)
                else:
                    content_paths = [Path(full_path, content) for content in contents]
                    contents = filesystem_manager.remove_mandatory_files(content_paths)
                    if not contents:
                        shutil.rmtree(full_path, ignore_errors=True)

        if os.path.exists(directory) and not os.listdir(directory):
            os.rmdir(directory)

    # class ProfileDirectories:
    #     def __init__(self, root_directory: Path) -> None:
    #         self.root_directory = Path(root_directory)
    #         self.metadata_directory = self.root_directory.joinpath("Metadata")
    #     def create_directories(self):
    #         self.root_directory.mkdir(exist_ok=True)

    class UserDirectories:
        def __init__(self) -> None:
            self.metadata_directory = Path()
            self.download_directory = Path()
            self.legacy_download_directories: list[Path] = []
            self.legacy_metadata_directories: list[Path] = []

        def find_legacy_directory(
            self,
            directory_type: Literal["metadata", "download"] = "metadata",
            api_type: str = "",
        ):
            match directory_type:
                case "metadata":
                    directories = self.legacy_metadata_directories
                case _:
                    directories = self.legacy_download_directories
            final_directory = directories[0]
            for directory in directories:
                for part in directory.parts:
                    if api_type in part:
                        return directory
            return final_directory

    async def walk(self, directory: Path):
        return await main_helper.walk(directory)


class FileManager:
    def __init__(self, directory_manager: DirectoryManager) -> None:
        self.files: list[Path] = []
        self.directory_manager = directory_manager

    async def set_default_files(
        self,
        prepared_metadata_format: ReformatItem,
        prepared_download_format: ReformatItem,
    ):
        self.files = []
        await self.update_files(prepared_metadata_format, "metadata_directory_format")
        await self.update_files(prepared_download_format, "file_directory_format")

    async def refresh_files(self):
        self.files = await self.directory_manager.walk(
            self.directory_manager.user.download_directory
        )
        return self.files

    async def update_files(self, reformatter: ReformatItem, format_key: str):
        directory_manager = self.directory_manager
        formatted_directory = reformatter.remove_non_unique(
            directory_manager, format_key
        )
        files: list[Path] = []
        files = await directory_manager.walk(formatted_directory)
        self.files.extend(files)
        return files

    async def add_file(self, filepath: Path):
        self.files.append(filepath)
        return True

    def remove_file(self, filepath: Path):
        self.files.remove(filepath)
        return True

    def delete_path(self, filepath: Path):
        if filepath.is_dir():
            filepath.rmdir()
        else:
            self.remove_file(filepath)
            filepath.unlink(missing_ok=True)
        return True

    async def cleanup(self):
        unique: set[Path] = set()
        await self.refresh_files()
        for valid_file in self.find_string_in_path("__drm__"):
            self.delete_path(valid_file)
            unique.add(valid_file.parent)
        for unique_file in unique:
            self.delete_path(unique_file)
        return True

    def find_string_in_path(self, string: str):
        valid_files: list[Path] = []
        for file in self.files:
            if string in file.as_posix():
                valid_files.append(file)
        return valid_files

    async def find_metadata_files(self, legacy_files: bool = True):
        new_list: list[Path] = []
        for filepath in self.files:
            if not legacy_files:
                if "__legacy_metadata__" in filepath.parts:
                    continue
            match filepath.suffix:
                case ".db":
                    red_list = ["thumbs.db"]
                    status = [x for x in red_list if x == filepath.name.lower()]
                    if status:
                        continue
                    new_list.append(filepath)
                case ".json":
                    new_list.append(filepath)
                case _:
                    pass
        return new_list


class FormatTypes:
    def __init__(self, site_settings: site_config_types) -> None:
        self.metadata_directory_format = site_settings.metadata_setup.directory_format
        self.file_directory_format = site_settings.download_setup.directory_format
        self.filename_format = site_settings.download_setup.filename_format

    def check_rules(self):
        """Checks for invalid filepath

        Returns:
            tuple(str,bool): Returns a string which explains invalid filepath format
        """
        bool_status = True
        wl = []
        invalid_list = []
        string = ""
        for key, _value in self:
            if key == "file_directory_format":
                bl = FormatAttributes()
                wl = [v for _k, v in bl.__dict__.items()]
                bl = bl.whitelist(wl)
                invalid_list = []
                for b in bl:
                    if b in self.file_directory_format.as_posix():
                        invalid_list.append(b)
            if key == "filename_format":
                bl = FormatAttributes()
                wl = [v for _k, v in bl.__dict__.items()]
                bl = bl.whitelist(wl)
                invalid_list = []
                for b in bl:
                    if b in self.filename_format.as_posix():
                        invalid_list.append(b)
            if key == "metadata_directory_format":
                wl = [
                    "{site_name}",
                    "{first_letter}",
                    "{model_id}",
                    "{profile_username}",
                    "{model_username}",
                ]
                bl = FormatAttributes().whitelist(wl)
                invalid_list: list[str] = []
                for b in bl:
                    if b in self.metadata_directory_format.as_posix():
                        invalid_list.append(b)
            if invalid_list:
                string += f"You cannot use {','.join(invalid_list)} in {key}. Use any from this list {','.join(wl)}"
                bool_status = False

        return string, bool_status

    def check_unique(self):
        values: list[str] = []
        unique = []
        new_format_copied = copy.deepcopy(self)
        option: dict[str, Any] = {}
        option["string"] = ""
        option["bool_status"] = True
        option["unique"] = new_format_copied
        f = FormatAttributes()
        for key, value in self:
            value: Path
            if key == "file_directory_format":
                unique = ["{media_id}", "{model_username}"]
                values = list(value.parts)
                option["unique"].file_directory_format = unique
            elif key == "filename_format":
                values = []
                unique = ["{media_id}", "{filename}"]
                for _key2, value2 in f:
                    if value2 in value.as_posix():
                        values.append(value2)
                option["unique"].filename_format = unique
            elif key == "metadata_directory_format":
                unique = ["{model_username}"]
                values = list(value.parts)
                option["unique"].metadata_directory_format = unique
            if key != "filename_format":
                e = [x for x in values if x in unique]
            else:
                e = [x for x in unique if x in values]
            if e:
                setattr(option["unique"], key, e)
            else:
                option[
                    "string"
                ] += f"{key} is a invalid format since it has no unique identifiers. Use any from this list {','.join(unique)}\n"
                option["bool_status"] = False
        return option

    def __iter__(self):
        for attr, value in self.__dict__.items():
            yield attr, value



# type: ignore
### posts.py ###

from sqlalchemy.orm import declarative_base

from ultima_scraper_collection.managers.database_manager.connections.sqlite.models.api_model import (
    ApiModel,
)
from ultima_scraper_collection.managers.database_manager.connections.sqlite.models.media_model import (
    TemplateMediaModel,
)

Base = declarative_base()


class api_table(ApiModel, Base):
    ApiModel.__tablename__ = "stories"


class TemplateMediaModel(TemplateMediaModel, Base):
    pass



# type: ignore

"""content

Revision ID: 1454e4d1c6b8
Revises: 37c4f2719d65
Create Date: 2023-02-12 23:17:29.239758

"""
import sqlalchemy as sa
from alembic import op
from sqlalchemy import Column, Integer

# revision identifiers, used by Alembic.
revision = "1454e4d1c6b8"
down_revision = "37c4f2719d65"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    conn = op.get_bind()

    meta = sa.MetaData()
    meta.reflect(bind=conn, only=("medias",))
    old_table = meta.tables["medias"]
    if "media_id_2" not in old_table.columns:
        with op.batch_alter_table("medias", recreate="always") as batch_op:
            batch_op.add_column(Column("media_id_2", Integer), insert_after="id")

    pass
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    pass
    # ### end Alembic commands ###



"""content

Revision ID: b791cf213df9
Revises: 5493253cc03c
Create Date: 2021-11-16 16:33:04.723478

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "b791cf213df9"
down_revision = "5493253cc03c"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "others",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("post_id", sa.Integer(), nullable=False),
        sa.Column("text", sa.String(), nullable=True),
        sa.Column("price", sa.Integer(), nullable=True),
        sa.Column("paid", sa.Integer(), nullable=True),
        sa.Column("archived", sa.Boolean(), nullable=True),
        sa.Column("created_at", sa.TIMESTAMP(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("post_id"),
    )
    op.create_table(
        "products",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("post_id", sa.Integer(), nullable=False),
        sa.Column("text", sa.String(), nullable=True),
        sa.Column("price", sa.Integer(), nullable=True),
        sa.Column("paid", sa.Integer(), nullable=True),
        sa.Column("archived", sa.Boolean(), nullable=True),
        sa.Column("created_at", sa.TIMESTAMP(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("post_id"),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("products")
    op.drop_table("others")
    # ### end Alembic commands ###



from ultima_scraper_db.databases.ultima_archive import merged_metadata
from ultima_scraper_db.databases.ultima_archive.api.client import UAClient
from ultima_scraper_db.databases.ultima_archive.database_api import ArchiveAPI
from ultima_scraper_db.managers.database_manager import Alembica

from ultima_scraper_collection.config import UltimaScraperCollectionConfig
from ultima_scraper_collection.projects.project_manager import Project


class UltimaArchiveProject(Project):
    async def init(self, config: UltimaScraperCollectionConfig):
        # We could pass a database manager instead of config
        db_info = config.settings.databases[0].connection_info.dict()
        ultima_archive_db = await super()._init_db(
            db_info, Alembica(generate=False, migrate=False), merged_metadata
        )
        self.ultima_archive_db_api = await ArchiveAPI(ultima_archive_db).init()
        self.fast_api = UAClient()
        UAClient.database_api = self.ultima_archive_db_api
        return self



# type: ignore

"""content

Revision ID: d2f2002f3c36
Revises: 1454e4d1c6b8
Create Date: 2023-02-13 22:40:57.202281

"""
import sqlalchemy as sa
from alembic import op
from sqlalchemy.orm import Session

# revision identifiers, used by Alembic.
revision = "d2f2002f3c36"
down_revision = "1454e4d1c6b8"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    invalid_conn = op.get_bind()
    database_url = str(invalid_conn.engine.url)
    conn = sa.create_engine(database_url)
    session = Session(bind=conn)
    res = session.execute(sa.text("SELECT id,media_id FROM medias;"))
    results = res.fetchall()
    meta = sa.MetaData()
    meta.reflect(bind=conn, only=("medias",))
    old_table = meta.tables["medias"]

    session = Session(bind=conn)
    for items in results:
        formatted = dict(items._mapping)
        (
            session.query(old_table)
            .filter(old_table.c.id == formatted["id"])
            .update({"media_id_2": formatted["media_id"]})
        )
        session.commit()
    with op.batch_alter_table("medias") as batch_op:
        batch_op.drop_column("media_id")
        batch_op.alter_column("media_id_2", new_column_name="media_id")
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    pass
    # ### end Alembic commands ###



from ultima_scraper_collection.managers.datascraper_manager.datascrapers.fansly import (
    FanslyDataScraper,
)
from ultima_scraper_collection.managers.datascraper_manager.datascrapers.onlyfans import (
    OnlyFansDataScraper,
)

datascraper_types = OnlyFansDataScraper | FanslyDataScraper



import socket
from typing import Sequence

import netifaces
from sqlalchemy import select
from ultima_scraper_db.databases.ultima_archive.database_api import ArchiveAPI
from ultima_scraper_db.databases.ultima_archive.schemas.management import (
    ServerModel,
    SiteModel,
)
from ultima_scraper_db.managers.database_manager import Database, Schema


class ServerManager:
    def __init__(self, ultima_archive_db_api: ArchiveAPI) -> None:
        self.ultima_archive_db_api = ultima_archive_db_api

    async def init(self, database: Database):
        def create_socket(socket_type: socket.SocketKind = socket.SOCK_DGRAM):
            temp_socket = socket.socket(socket.AF_INET, socket_type)
            temp_socket.connect(("8.8.8.8", 80))  # Connecting to Google's DNS server
            return temp_socket

        def get_local_ip():
            # Create a temporary connection to a remote server to retrieve the local IP address
            temp_socket = create_socket()
            local_ip = temp_socket.getsockname()[0]
            temp_socket.close()
            return local_ip

        def mac_for_ip(ip: str) -> str | None:
            "Returns a list of MACs for interfaces that have given IP, returns None if not found"
            for i in netifaces.interfaces():  # type: ignore
                addrs = netifaces.ifaddresses(i)  # type: ignore
                try:
                    if_mac: str | None = addrs[netifaces.AF_LINK][0]["addr"]  # type: ignore
                    if_ip: str | None = addrs[netifaces.AF_INET][0]["addr"]  # type: ignore
                except (IndexError, KeyError):  # ignore ifaces that dont have MAC or IP
                    if_mac = if_ip = None
                if if_ip == ip:
                    return if_mac  # type: ignore
            return None

        self.management_db = database.schemas["management"]
        db_sites = await self.management_db.session.scalars(select(SiteModel))
        db_sites = db_sites.all()
        self.reset = False
        if not db_sites:
            # Need to add a create or update for additional sites
            from ultima_scraper_db.databases.ultima_archive.schemas.management import (
                default_sites,
            )

            for site in default_sites:
                self.management_db.session.add(site)
            await self.management_db.session.commit()
            db_sites = await self.management_db.session.scalars(select(SiteModel))
            db_sites = db_sites.all()
            self.reset = True
        private_ip = get_local_ip()
        mac_address = mac_for_ip(private_ip)
        # public_ip = requests.get("https://checkip.amazonaws.com/").text.strip()
        self.db_sites: Sequence[SiteModel] = db_sites
        self.ip_address = private_ip

        db_servers = await self.management_db.session.scalars(select(ServerModel))
        db_servers = db_servers.all()
        if not db_servers:
            default_server = ServerModel(
                name="home", ip=self.ip_address, mac_address=mac_address
            )
            self.management_db.session.add(default_server)
            await self.management_db.session.commit()
        active_server = await self.management_db.session.scalars(
            select(ServerModel).where(
                (ServerModel.ip == self.ip_address)
                & (ServerModel.mac_address == mac_address)
            )
        )
        self.active_server = active_server.one()
        self.site_schemas: dict[str, Schema] = {}
        for db_site in self.db_sites:
            site_schema_api = self.ultima_archive_db_api.get_site_api(db_site.db_name)
            self.site_schemas[site_schema_api.schema.name] = site_schema_api.schema
        return self

    async def resolve_site_schema(self, value: str):
        return self.site_schemas[value]

    async def resolve_db_site(self, value: str):
        return [x for x in self.db_sites if x.db_name == value][0]

    async def find_site_api(self, name: str):
        return self.ultima_archive_db_api.site_apis[name]

    def get_server_id(self):
        return self.active_server.id



from __future__ import annotations

import copy
from itertools import product
from typing import Any, Optional

import ultima_scraper_api
from sqlalchemy import and_, or_, select
from sqlalchemy.orm import joinedload
from tqdm.asyncio import tqdm_asyncio
from ultima_scraper_api.apis.onlyfans.classes.auth_model import OnlyFansAuthModel
from ultima_scraper_collection.config import site_config_types
from ultima_scraper_collection.managers.content_manager import (
    ContentManager,
    MediaManager,
)
from ultima_scraper_collection.managers.download_manager import DownloadManager
from ultima_scraper_collection.managers.filesystem_manager import FilesystemManager
from ultima_scraper_collection.managers.metadata_manager.metadata_manager import (
    MediaMetadata,
    MetadataManager,
)
from ultima_scraper_collection.managers.server_manager import ServerManager
from ultima_scraper_db.databases.ultima_archive.schemas.templates.site import (
    FilePathModel as DBFilePathModel,
)
from ultima_scraper_db.databases.ultima_archive.schemas.templates.site import (
    MediaModel as DBMediaModel,
)
from ultima_scraper_db.databases.ultima_archive.schemas.templates.site import (
    MessageModel as DBMessageModel,
)
from ultima_scraper_db.databases.ultima_archive.schemas.templates.site import UserModel
from ultima_scraper_db.databases.ultima_archive.schemas.templates.site import (
    UserModel as DBUserModel,
)
from ultima_scraper_renamer.reformat import ReformatManager

auth_types = ultima_scraper_api.auth_types
user_types = ultima_scraper_api.user_types
message_types = ultima_scraper_api.message_types
error_types = ultima_scraper_api.error_types
subscription_types = ultima_scraper_api.subscription_types
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ultima_scraper_api.managers.session_manager import AuthedSession
    from ultima_scraper_collection.managers.datascraper_manager.datascrapers.fansly import (
        FanslyDataScraper,
    )
    from ultima_scraper_collection.managers.datascraper_manager.datascrapers.onlyfans import (
        OnlyFansDataScraper,
    )
    from ultima_scraper_collection.managers.metadata_manager.metadata_manager import (
        ContentMetadata,
    )

    datascraper_types = OnlyFansDataScraper | FanslyDataScraper


class download_session(tqdm_asyncio):
    def start(
        self,
        unit: str = "B",
        unit_scale: bool = True,
        miniters: int = 1,
        tsize: int = 0,
    ):
        self.unit = unit
        self.unit_scale = unit_scale
        self.miniters = miniters
        self.total = 0
        self.colour = "Green"
        if tsize:
            tsize = int(tsize)
            self.total += tsize

    def update_total_size(self, tsize: Optional[int]):
        if tsize:
            tsize = int(tsize)
            self.total += tsize


async def find_earliest_non_downloaded_message(
    user: user_types, datascraper: "datascraper_types"
):
    authed = user.get_authed()
    site_api = datascraper.server_manager.ultima_archive_db_api.get_site_api(
        authed.get_api().site_name
    )
    temp_db_messages = await site_api.schema.session.scalars(
        select(DBMessageModel)
        .where(
            or_(
                and_(
                    DBMessageModel.user_id == authed.id,
                    DBMessageModel.receiver_id == user.id,
                ),
                and_(
                    DBMessageModel.user_id == user.id,
                    DBMessageModel.receiver_id == authed.id,
                ),
            )
        )
        .order_by(DBMessageModel.id.desc())
        .options(
            joinedload(DBMessageModel.media)
            .joinedload(DBMediaModel.filepaths)
            .joinedload(DBFilePathModel.message)
        )
    )

    db_messages = temp_db_messages.unique().all()
    earliest_non_downloaded_message = None

    for db_message in db_messages:
        all_media_downloaded = False
        for media in db_message.media:
            for filepath in media.filepaths:
                if not filepath.message:
                    continue
                if filepath.downloaded:
                    all_media_downloaded = True
                    break

        if not all_media_downloaded:
            if (
                earliest_non_downloaded_message is None
                or db_message.created_at < earliest_non_downloaded_message.created_at
            ):
                earliest_non_downloaded_message = db_message
    return earliest_non_downloaded_message


class StreamlinedDatascraper:
    def __init__(
        self, datascraper: datascraper_types, server_manager: ServerManager
    ) -> None:
        self.datascraper = datascraper
        self.filesystem_manager = FilesystemManager()
        self.media_types = self.datascraper.api.MediaTypes()
        self.user_list: set[user_types] = set()
        self.metadata_manager_users: dict[int, MetadataManager] = {}
        self.server_manager: ServerManager = server_manager
        self.content_managers: dict[int, ContentManager] = {}
        self.media_managers: dict[int, MediaManager] = {}

    def resolve_content_manager(self, user: user_types):
        content_manager = self.content_managers.get(user.id)
        authed = user.get_authed()
        if content_manager:
            if content_manager.authed.id != authed.id:
                content_manager = ContentManager(authed)
        else:
            content_manager = ContentManager(authed)
        self.content_managers[user.id] = content_manager
        return content_manager

    def create_media_manager(self, user: user_types):
        if user.id not in self.media_managers:
            self.media_managers[user.id] = MediaManager()
        return self.media_managers[user.id]

    def get_archive_db_api(self):
        return self.server_manager.ultima_archive_db_api

    async def configure_datascraper_jobs(self):
        api = self.datascraper.api
        site_config = self.datascraper.site_config
        available_jobs = site_config.jobs.scrape
        option_manager = self.datascraper.option_manager
        performer_options = option_manager.performer_options
        assert option_manager.subscription_options
        valid_user_list: set[user_types] = set(
            option_manager.subscription_options.final_choices
        )
        scraping_subscriptions = site_config.jobs.scrape.subscriptions
        identifiers = []
        if performer_options:
            identifiers = performer_options.return_auto_choice()
        if not available_jobs.subscriptions:
            for authed in api.auths.values():
                authed.subscriptions = []
        if available_jobs.messages:
            chat_users: list[user_types] = []
            if identifiers:
                for authed in api.auths.values():
                    for identifier in identifiers:
                        chat_user = await authed.get_user(identifier)
                        if isinstance(chat_user, user_types):
                            chat_users.append(chat_user)
            else:
                chat_users = await self.get_chat_users()
            [
                user.scrape_whitelist.append("Messages")
                for user in chat_users
                if not user.is_subscribed() or not scraping_subscriptions
            ]
            [valid_user_list.add(x) for x in chat_users]

        if available_jobs.paid_contents:
            for authed in self.datascraper.api.auths.values():
                paid_contents = await authed.get_paid_content()
                if not isinstance(paid_contents, error_types):
                    for paid_content in paid_contents:
                        author = paid_content.get_author()
                        if identifiers:
                            found = await author.match_identifiers(identifiers)
                            if not found:
                                continue
                        if author:
                            performer = authed.find_user(
                                identifier=author.id,
                            )
                            if performer:
                                performer.job_whitelist.append("PaidContents")
                                performer.scrape_whitelist.clear()
                                valid_user_list.add(performer)
        from ultima_scraper_api.apis.fansly.classes.user_model import (
            create_user as FYUserModel,
        )

        for user in valid_user_list:
            if isinstance(user, FYUserModel) and user.following:
                user.scrape_whitelist.clear()
                pass
            pass
        # Need to filter out own profile with is_performer,etc
        final_valid_user_set = {
            user
            for user in valid_user_list
            if user.username not in user.get_authed().blacklist
        }

        self.user_list = final_valid_user_set
        return final_valid_user_set

    # Prepares the API links to be scraped
    async def scrape_vault(
        self, user: user_types, db_user: UserModel, content_type: str
    ):
        current_job = user.get_current_job()
        if not current_job:
            return
        authed: auth_types = user.get_authed()
        site_config = self.datascraper.site_config
        if isinstance(authed, OnlyFansAuthModel) and user.is_authed_user():
            vault = await authed.get_vault_lists()
            vault_item = vault.resolve(name=content_type)
            assert vault_item
            vault_item_medias = await vault_item.get_medias()
            media_metadatas: list[MediaMetadata] = []
            for vault_item_media in vault_item_medias:
                content_manager = self.resolve_content_manager(user)
                media_metadata = MediaMetadata(
                    vault_item_media["id"],
                    vault_item_media["type"],
                    content_manager=content_manager,
                )
                media_metadata.raw_extractor(user, vault_item_media)
                reformat_manager = ReformatManager(authed, self.filesystem_manager)
                reformat_item = reformat_manager.prepare_reformat(media_metadata)
                file_directory = reformat_item.reformat(
                    site_config.download_setup.directory_format
                )
                reformat_item.directory = file_directory
                file_path = reformat_item.reformat(
                    site_config.download_setup.filename_format
                )
                media_metadata.directory = file_directory
                media_metadata.filename = file_path.name
                media_metadatas.append(media_metadata)
        current_job.done = True

    async def prepare_scraper(
        self,
        user: user_types,
        metadata_manager: MetadataManager,
        content_type: str,
        master_set: list[Any] = [],
    ):
        authed = user.get_authed()
        current_job = user.get_current_job()
        if not current_job:
            return
        temp_master_set: list[Any] = copy.copy(master_set)
        if not temp_master_set and not current_job.ignore:
            match content_type:
                case "Stories":
                    temp_master_set.extend(await self.datascraper.get_all_stories(user))
                    pass
                case "Posts":
                    temp_master_set = await self.datascraper.get_all_posts(user)
                case "Messages":
                    db_message = await find_earliest_non_downloaded_message(
                        user, self.datascraper
                    )
                    cutoff_id = db_message.id if db_message else None
                    temp_master_set = await user.get_messages(cutoff_id=cutoff_id)
                case "Chats":
                    pass
                case "Highlights":
                    pass
                case "MassMessages":
                    if isinstance(authed, OnlyFansAuthModel):
                        if user.is_authed_user():
                            mass_message_stats = await authed.get_mass_message_stats()
                            temp_master_set = []
                            for mass_message_stat in mass_message_stats:
                                mass_message = (
                                    await mass_message_stat.get_mass_message()
                                )
                                temp_master_set.append(mass_message)
                        else:
                            db_message = await find_earliest_non_downloaded_message(
                                user, self.datascraper
                            )
                            cutoff_id = db_message.id if db_message else None
                            mass_messages = await user.get_mass_messages(
                                message_cutoff_id=cutoff_id
                            )
                            temp_master_set.extend(mass_messages)
                case _:
                    raise Exception(f"{content_type} is an invalid choice")
        # Adding paid content and removing duplicates by id
        if isinstance(user, ultima_scraper_api.onlyfans_classes.user_model.create_user):
            for paid_content in await user.get_paid_contents(content_type):
                temp_master_set.append(paid_content)
            pass
            temp_master_set = list(
                {getattr(obj, "id"): obj for obj in temp_master_set}.values()
            )
        await self.process_scraped_content(
            temp_master_set, content_type, user, metadata_manager
        )
        current_job.done = True

    async def process_scraped_content(
        self,
        master_set: list[dict[str, Any]],
        api_type: str,
        subscription: user_types,
        metadata_manager: MetadataManager,
    ):
        if not master_set:
            return False
        # if  api_type != "Posts":
        #     return
        authed = subscription.get_authed()
        unrefined_set = []
        with authed.get_pool() as pool:
            print(f"Processing Scraped {api_type}")
            tasks = pool.starmap(
                self.datascraper.media_scraper,
                product(
                    master_set,
                    [subscription],
                    [api_type],
                ),
            )
            settings = {"colour": "MAGENTA"}
            unrefined_set: list[dict[str, Any]] = await tqdm_asyncio.gather(
                *tasks, **settings
            )
            new_metadata = metadata_manager.merge_content_and_directories(unrefined_set)
            final_content, _final_directories = new_metadata
            if new_metadata:
                new_metadata_content = final_content
                content_manager = self.resolve_content_manager(subscription)
                content_manager.set_content(api_type, new_metadata_content)
                if new_metadata_content:
                    pass
            else:
                print(f"No {api_type} found.")
        return True

    # Downloads scraped content
    async def prepare_downloads(
        self, performer: user_types, db_performer: DBUserModel, api_type: str
    ):
        site_db_api = self.server_manager.ultima_archive_db_api.find_site_api(
            self.datascraper.api.site_name
        )
        current_job = performer.get_current_job()
        global_settings = performer.get_api().get_global_settings()
        filesystem_manager = self.datascraper.filesystem_manager
        performer_directory_manager = filesystem_manager.get_directory_manager(
            performer.id
        )
        filesystem_manager = self.datascraper.filesystem_manager
        content_manager = self.resolve_content_manager(performer)
        db_medias = db_performer.content_manager.get_media_manager().medias
        final_download_set: set[MediaMetadata] = set()
        for db_media in db_medias.values():
            content_info = None
            if api_type == "Uncategorized":
                await db_media.awaitable_attrs.content_media_assos
                if db_media.content_media_assos:
                    continue
                if len(db_media.filepaths) > 1:
                    continue
            else:
                db_content = await db_media.find_content(api_type)
                if not db_content:
                    continue
                content_info = (db_content.id, api_type)
            db_filepath = db_media.find_filepath(content_info)
            if db_filepath:
                if api_type == "Uncategorized":
                    media_metadata = content_manager.media_manager.medias.get(
                        db_media.id
                    )
                else:
                    media_metadata = content_manager.find_media(
                        category=api_type, media_id=db_media.id
                    )
                if media_metadata:
                    media_metadata.__db_media__ = db_media
                    final_download_set.add(media_metadata)
        total_media_count = len(final_download_set)
        download_media_count = 0
        directory = performer_directory_manager.user.download_directory
        if final_download_set:
            string = "Processing Download:\n"
            string += f"Name: {performer.username} | Type: {api_type} | Downloading: {download_media_count} | Total: {total_media_count} | Directory: {directory}\n"
            print(string)
        download_manager = DownloadManager(
            performer.get_authed(),
            filesystem_manager,
            final_download_set,
            global_settings.tools.reformatter.active,
        )
        await download_manager.bulk_download()
        await site_db_api.schema.session.commit()
        if final_download_set:
            pass
        if current_job:
            current_job.done = True

    async def manage_subscriptions(
        self,
        authed: auth_types,
        identifiers: list[int | str] = [],
        refresh: bool = True,
    ):
        temp_subscriptions: list[subscription_types] = []
        results = await self.datascraper.get_all_subscriptions(
            authed, identifiers, refresh
        )
        site_settings = authed.api.get_site_settings()
        if not site_settings:
            return temp_subscriptions
        results.sort(key=lambda x: x.user.is_me(), reverse=True)
        for result in results:
            temp_subscriptions.append(result)
        authed.subscriptions = temp_subscriptions
        return authed.subscriptions

    async def account_setup(
        self,
        auth: auth_types,
        site_config: site_config_types,
        identifiers: list[int | str] | list[str] = [],
    ) -> tuple[bool, list[subscription_types]]:
        status = False
        subscriptions: list[subscription_types] = []

        if auth.is_authed() and site_config:
            authed = auth
            # metadata_filepath = (
            #     authed.directory_manager.profile.metadata_directory.joinpath(
            #         "Mass Messages.json"
            #     )
            # )
            # if authed.isPerformer:
            #     imported = main_helper.import_json(metadata_filepath)
            #     if "auth" in imported:
            #         imported = imported["auth"]
            #     mass_messages = await authed.get_mass_messages(resume=imported)
            #     if mass_messages:
            #         main_helper.export_json(mass_messages, metadata_filepath)
            authed.blacklist = await authed.get_blacklist(site_config.blacklists)
            if identifiers or site_config.jobs.scrape.subscriptions:
                subscriptions.extend(
                    await self.manage_subscriptions(
                        authed, identifiers=identifiers  # type: ignore
                    )
                )
            status = True
        return status, subscriptions

    async def get_chat_users(self):
        chat_users: list[user_types] = []
        for authed in self.datascraper.api.auths:
            chats = await authed.get_chats()
            for chat in chats:
                username: str = chat["withUser"].username
                subscription = await authed.get_subscription(identifier=username)
                if not subscription:
                    subscription = chat["withUser"]
                    chat_users.append(subscription)
        return chat_users

    async def get_performer(self, authed: auth_types, db_performer: DBUserModel):
        if authed.id == db_performer.id:
            performer = authed.user
        else:
            subscriptions = await authed.get_subscriptions(
                identifiers=[db_performer.id]
            )
            if not subscriptions:
                paid_contents = await authed.get_paid_content(
                    performer_id=db_performer.id
                )
                if not paid_contents:
                    # performer = await authed.get_user(db_performer.id)
                    # if not performer.subscribed_by_data:
                    #     return None
                    return None
                else:
                    performer = paid_contents[0].get_author()
            else:
                performer = subscriptions[0].user
                if not performer.is_subscribed():
                    paid_contents = await authed.get_paid_content(
                        performer_id=db_performer.id
                    )
        if isinstance(
            performer, ultima_scraper_api.onlyfans_classes.user_model.create_user
        ):
            if performer.is_blocked:
                await performer.unblock()
        return performer



from pathlib import Path

from pydantic import BaseModel, StrictBool, StrictInt, StrictStr
from ultima_scraper_api.config import FanslyAPIConfig
from ultima_scraper_api.config import GlobalAPI as USAGlobalAPI
from ultima_scraper_api.config import OnlyFansAPIConfig, Settings
from ultima_scraper_api.config import Sites as USASites
from ultima_scraper_api.config import UltimaScraperAPIConfig


class Jobs(BaseModel):
    class Scrape(BaseModel):
        subscriptions: bool = True
        messages: bool = True
        paid_contents: bool = True

    class Metadata:
        content: bool = True
        comments: bool = True

    scrape: Scrape = Scrape()
    metadata: Scrape = Scrape()


class Directory(BaseModel):
    path: Path | None = None
    minimum_space: int = -1
    store: bool = True
    overflow: bool = True


class GlobalXPathSetup(BaseModel):
    directories: list[Directory] = [Directory(path=Path("__user_data__").absolute())]
    directory_format: Path = Path()


class DownloadPathSetup(GlobalXPathSetup):
    filename_format: Path = Path()
    text_length: int = 255
    date_format: str = "%Y-%m-%d"
    overwrite_files: bool = True


class Trash(BaseModel):
    cleanup: bool = True


class ToolSettings(BaseModel):
    active: bool = True


class Renamer(ToolSettings):
    pass


class Reformatter(ToolSettings):
    pass


class Downloader(ToolSettings):
    pass


class SSHConnection(BaseModel):
    username: str | None = None
    private_key_filepath: Path | None = None
    private_key_password: str | None = None
    host: str | None = None
    port: int = 22


class DatabaseInfo(BaseModel):
    name: str = "ultima_archive"
    host: str = "localhost"
    port: int = 5432
    username: str | None = None
    password: str | None = None
    ssh: SSHConnection = SSHConnection()


class Database(BaseModel):
    connection_info: DatabaseInfo = DatabaseInfo()
    main: bool = True
    active: bool = True


class Tools(BaseModel):
    renamer: Renamer = Renamer()
    reformatter: Reformatter = Reformatter()
    downloader: Downloader = Downloader()


auto_types = list[int | str] | StrictInt | StrictStr | StrictBool | None


class GlobalAPI(USAGlobalAPI):
    auto_profile_choice: auto_types = None
    auto_performer_choice: auto_types = None
    auto_content_choice: auto_types = None
    auto_media_choice: auto_types = None
    jobs = Jobs()
    metadata_setup = GlobalXPathSetup()
    metadata_setup.directory_format = (
        "{site_name}/{first_letter}/{model_username}/Metadata"
    )  # type: ignore
    download_setup = DownloadPathSetup()
    download_setup.directory_format = (
        "{site_name}/{first_letter}/{model_username}/{api_type}/{value}/{media_type}"
    )  # type: ignore
    download_setup.filename_format = "{filename}.{ext}"  # type: ignore
    blacklists: list[str] = []


class Sites(USASites):
    class OnlyFansAPIConfig(OnlyFansAPIConfig, GlobalAPI):
        pass

    class FanslyAPIConfig(FanslyAPIConfig, GlobalAPI):
        pass

    onlyfans: OnlyFansAPIConfig = OnlyFansAPIConfig(auto_content_choice=True)
    fansly: FanslyAPIConfig = FanslyAPIConfig()


site_config_types = Sites.OnlyFansAPIConfig | Sites.FanslyAPIConfig


class UltimaScraperCollectionConfig(UltimaScraperAPIConfig):
    class Settings(Settings):
        auto_site_choice: str = ""
        databases: list[Database] = [Database()]
        tools: Tools = Tools()
        trash = Trash()
        infinite_loop: bool = False
        exit_on_completion: bool = True

        def get_main_database(self):
            return [x for x in self.databases if x.main][0]

    settings: Settings = Settings()
    site_apis: Sites = Sites()

    def get_site_config(self, site_name: str):
        return getattr(self.site_apis, site_name.lower())



import mimetypes
import re
from datetime import datetime
from pathlib import Path
from typing import Any, TypedDict
from urllib.parse import ParseResult, urlparse

import ultima_scraper_api
from sqlalchemy import inspect
from ultima_scraper_api.apis.onlyfans import preview_url_picker, url_picker
from ultima_scraper_api.apis.onlyfans.classes.mass_message_model import MassMessageModel
from ultima_scraper_api.helpers import main_helper
from ultima_scraper_collection.managers.content_manager import ContentManager
from ultima_scraper_collection.managers.database_manager.connections.sqlite.sqlite_database import (
    DBCollection,
    SqliteDatabase,
)
from ultima_scraper_collection.managers.database_manager.database_manager import (
    DatabaseManager,
)
from ultima_scraper_collection.managers.filesystem_manager import FilesystemManager
from ultima_scraper_db.databases.ultima_archive.schemas.templates.site import (
    MediaModel as DBMediaModel,
)
from ultima_scraper_db.databases.ultima_archive.schemas.templates.site import (
    MessageModel as DBMessageModel,
)
from ultima_scraper_db.databases.ultima_archive.schemas.templates.site import (
    PostModel as DBPostModel,
)
from ultima_scraper_db.databases.ultima_archive.schemas.templates.site import (
    StoryModel as DBStoryModel,
)

api_types = ultima_scraper_api.api_types
user_types = ultima_scraper_api.user_types
content_types = ultima_scraper_api.content_types


class DBContentExtractor:
    def __init__(
        self, content_item: DBMessageModel | DBPostModel | DBStoryModel
    ) -> None:
        self.item = content_item
        self.__api__: api_types | None = None

    def get_id(self):
        return self.item.id

    def get_user_id(self):
        return self.item.user_id

    def get_queue_id(self):
        return self.item.queue_id

    def get_text(self):
        if isinstance(self.item, DBStoryModel):
            return None
        return self.item.text

    def get_preview_ids(self):
        value: list[int] = []
        return value

    def resolve_archived(self):
        return False

    def get_date(self):
        return self.item.created_at

    def get_medias(self, content_metadata: "ContentMetadata"):
        final_assets: list[MediaMetadata] = []
        for media_item in self.item.media:
            if not media_item.category:
                from urllib.parse import urlparse

                if not media_item.url:
                    continue
                parsed_url = urlparse(media_item.url)
                path_ = parsed_url.path
                type_, _encoding = mimetypes.guess_type(path_)
                media_type = None
                if type_:
                    media_type, _subtype = type_.split("/")
                assert media_type and self.__api__
                true_media_type = self.__api__.MediaTypes().find_by_value(media_type)
                pass
            else:
                true_media_type = media_item.category

            new_asset = MediaMetadata(
                media_item.id,
                true_media_type,
                content_manager=content_metadata.content_manager,
                content_metadata=content_metadata,
                created_at=content_metadata.created_at,
            )
            new_asset.urls = [media_item.url]
            filepath = media_item.find_filepath(
                content_metadata.content_id, content_metadata.api_type
            )
            if filepath:
                pathed_filepath = Path(filepath.filepath)
                new_asset.directory = pathed_filepath.parent
                new_asset.filename = pathed_filepath.name
            else:
                pass
            new_asset.size = media_item.size
            final_assets.append(new_asset)
        return final_assets

    def resolve_paid(self):
        if not hasattr(self.item, "paid"):
            return None
        if isinstance(self.item, DBPostModel | DBMessageModel):
            return self.item.paid

    def get_receiver_id(self):
        if isinstance(self.item, DBMessageModel):
            return self.item.receiver_id


class ApiExtractor:
    def __init__(self, item: content_types) -> None:
        self.item = item

    def get_id(self):
        return self.item.id

    def get_user_id(self):
        return self.item.get_author().id

    def get_queue_id(self):
        if isinstance(self.item, ultima_scraper_api.message_types):
            return self.item.queue_id

    def get_is_from_queue(self):
        if isinstance(self.item, ultima_scraper_api.message_types):
            return self.item.is_from_queue

    def get_text(self):
        if isinstance(self.item, ultima_scraper_api.post_types):
            final_text = self.item.rawText or self.item.text
        elif isinstance(self.item, ultima_scraper_api.message_types):
            final_text = self.item.text
        elif isinstance(self.item, MassMessageModel):
            final_text = self.item.raw_text or self.item.text
        else:
            final_text = ""
        return final_text

    def get_preview_ids(self):
        if isinstance(self.item, ultima_scraper_api.post_types):
            final_preview_ids = self.item.preview or self.item.preview_ids
        elif isinstance(self.item, ultima_scraper_api.message_types):
            final_preview_ids = self.item.previews
        elif isinstance(self.item, MassMessageModel):
            final_preview_ids = self.item.previews
        else:
            final_preview_ids = []
        return final_preview_ids

    def get_date(self):
        temp_date = self.item.created_at
        assert temp_date
        if isinstance(temp_date, str):
            try:
                temp_date = float(temp_date)
            except:
                pass
        default_date = datetime.now()
        if temp_date == "-001-11-30T00:00:00+00:00":
            date_object = default_date
        else:
            if not temp_date:
                temp_date = default_date
            if isinstance(temp_date, int):
                timestamp = float(temp_date)
                date_object = datetime.fromtimestamp(timestamp)
            elif isinstance(temp_date, float):
                date_object = datetime.fromtimestamp(temp_date)
            elif isinstance(temp_date, str):
                date_object = datetime.fromisoformat(temp_date)
            else:
                date_object = temp_date
        return date_object

    def get_medias(self, content_metadata: "ContentMetadata"):
        final_assets: list[MediaMetadata] = []
        for asset_metadata in self.item.media:
            raw_media_type = asset_metadata["type"]
            if "mimetype" in asset_metadata:
                raw_media_type: str = asset_metadata["mimetype"].split("/")[0]
            author = self.item.get_author()
            media_type = author.get_api().MediaTypes().find_by_value(raw_media_type)
            if "createdAt" in asset_metadata:
                if isinstance(asset_metadata["createdAt"], str):
                    media_created_at = datetime.fromisoformat(
                        asset_metadata["createdAt"]
                    )
                else:
                    media_created_at = datetime.fromtimestamp(
                        asset_metadata["createdAt"]
                    )
            else:
                media_created_at = content_metadata.created_at
            new_asset = MediaMetadata(
                asset_metadata["id"],
                media_type,
                content_manager=content_metadata.content_manager,
                content_metadata=content_metadata,
                created_at=media_created_at,
            )
            new_asset.author = author
            new_asset.user_id = author.id
            media_manager = content_metadata.content_manager.media_manager
            if new_asset.id:
                media_manager.medias[new_asset.id] = new_asset
            else:
                media_manager.invalid_medias.append(new_asset)
            main_url = self.item.url_picker(asset_metadata)
            preview_url = self.item.preview_url_picker(asset_metadata)
            authed = author.get_authed()
            if authed.drm:
                new_asset.drm = bool(authed.drm.has_drm(asset_metadata))
            new_asset.urls = []
            matches = ["us", "uk", "ca", "ca2", "de"]
            for url in [main_url, preview_url].copy():
                if url:
                    if url.hostname:
                        subdomain = url.hostname.split(".")[1]
                        if any(subdomain in nm for nm in matches):
                            subdomain = url.hostname.split(".")[1]
                            if "upload" in subdomain:
                                continue
                            if "convert" in subdomain:
                                continue
                    new_asset.urls.append(url.geturl())

            if new_asset.id and new_asset.id in content_metadata.preview_media_ids:
                new_asset.preview = True
            new_asset.__raw__ = asset_metadata
            final_assets.append(new_asset)
        return final_assets

    def resolve_archived(self):
        archived = False
        if isinstance(self.item, ultima_scraper_api.post_types):
            archived = self.item.isArchived
        return archived

    def resolve_paid(self):
        if hasattr(self.item, "price"):
            if all(media["canView"] for media in self.item.media):
                return True
            return False
        else:
            return None

    def get_receiver_id(self):
        if isinstance(self.item, ultima_scraper_api.message_types):
            return self.item.get_receiver().id


class ContentMetadata:
    def __init__(
        self,
        content_id: int,
        api_type: str,
        content_manager: ContentManager,
    ) -> None:
        self.content_id = content_id
        self.user_id: int | None = None
        self.receiver_id: int | None = None
        self.queue_id: int | None = None
        self.text: str | None = None
        self.preview_media_ids: list[int] | list[dict[str, Any]] = []
        self.archived: bool = False
        self.medias: list[MediaMetadata] = []
        self.api_type = api_type
        self.price: float | None = None
        self.paid: bool | None = False
        self.deleted: bool = False
        self.is_from_queue: bool = False
        self.__raw__: Any | None = None
        self.__soft__: Any = None
        self.__db_content__: DBStoryModel | DBPostModel | DBMessageModel | None = None
        self.__legacy__ = False
        self.content_manager = content_manager

    async def resolve_extractor(self, result: ApiExtractor | DBContentExtractor):
        self.content_id = result.get_id()
        self.user_id = result.get_user_id()
        self.receiver_id = result.get_receiver_id()
        self.queue_id = result.get_queue_id()
        self.is_from_queue = bool(result.get_is_from_queue())
        self.text = result.get_text()
        self.preview_media_ids = result.get_preview_ids()
        self.archived = result.resolve_archived()
        self.price = getattr(result.item, "price", 0) or 0
        self.paid = result.resolve_paid()
        self.deleted = False
        self.created_at: datetime = result.get_date()
        self.medias: list[MediaMetadata] = result.get_medias(self)
        self.__raw__: Any | None = None
        self.__soft__ = result.item
        self.__media_types__ = None

    def find_media(self, media_id: int | None = None, urls: list[str] = []):
        for asset in self.medias:
            if asset.id == media_id:
                return asset
        for asset in self.medias:
            for url in urls:
                url = urlparse(url)
                found_url = asset.find_by_url(url)
                if found_url:
                    return asset

    def find_media_metadata_by_id(self, media_id: int):
        for asset in self.medias:
            if asset.id == media_id:
                return asset

    def find_media_by_url(self, urls: list[str]):
        for asset in self.medias:
            for url in urls:
                url = urlparse(url)
                found_url = asset.find_by_url(url)
                if found_url:
                    return asset

    def get_mass_message_stat(self):
        if isinstance(self.__soft__, MassMessageModel):
            return self.__soft__.mass_message_stat


class MediaMetadata:
    def __init__(
        self,
        media_id: int,
        media_type: str,
        content_manager: ContentManager,
        content_metadata: ContentMetadata | None = None,
        urls: list[str] = [],
        preview: bool = False,
        created_at: datetime | None = None,
        drm: bool = False,
    ) -> None:
        self.id = int(media_id) if media_id is not None else None
        self.media_type = media_type
        self.urls: list[str] = urls
        self.preview = preview
        self.directory: Path | None = None
        self.filename: str | None = None
        self.size = 0
        self.linked = None
        self.drm = drm
        self.key: str = ""
        self.created_at = created_at
        self.__content_metadata__ = content_metadata
        self.__db_media__: DBMediaModel | None = None
        self.content_manager = content_manager
        assert self.id
        self.content_manager.media_manager.medias[self.id] = self

    def get_author(self):
        return self.author

    def raw_extractor(self, author: user_types, raw_media: dict[str, Any]):
        self.author = author
        self.user_id = author.id
        assert self.id
        raw_media_type = raw_media["type"]
        if "mimetype" in raw_media:
            raw_media_type: str = raw_media["mimetype"].split("/")[0]
        self.media_type = author.get_api().MediaTypes().find_by_value(raw_media_type)
        self.drm = bool(author.get_authed().drm.has_drm(raw_media))
        main_url = self.url_picker(raw_media)
        preview_url = self.preview_url_picker(raw_media)
        matches = ["us", "uk", "ca", "ca2", "de"]
        self.urls = []
        for url in [main_url, preview_url].copy():
            if url:
                if url.hostname:
                    subdomain = url.hostname.split(".")[1]
                    if any(subdomain in nm for nm in matches):
                        subdomain = url.hostname.split(".")[1]
                        if "upload" in subdomain:
                            continue
                        if "convert" in subdomain:
                            continue
                self.urls.append(url.geturl())
        if (
            self.__content_metadata__
            and self.id in self.__content_metadata__.preview_media_ids
        ):
            self.preview = True

        if isinstance(raw_media["createdAt"], str):
            media_created_at = datetime.fromisoformat(raw_media["createdAt"])
        else:
            media_created_at = datetime.fromtimestamp(raw_media["createdAt"])
        self.created_at = media_created_at
        self.__raw__ = raw_media

    def url_picker(self, media_item: dict[str, Any], video_quality: str = ""):
        return url_picker(self.author, media_item, video_quality)

    def preview_url_picker(self, media_item: dict[str, Any]):
        return preview_url_picker(media_item)

    def find_by_url(self, url: ParseResult):
        for media_url in self.urls:
            media_url = urlparse(media_url)
            url_path = url.path
            if "Protected" in url.path:
                index = url.path.index("/files")
                url_path = url.path[index:]
            if media_url.path == url_path:
                return self

    def get_filepath(self):
        assert self.directory and self.filename
        return Path(self.directory, self.filename)

    def get_content_metadata(self):
        return self.__content_metadata__


class MetadataManager:
    def __init__(
        self,
        subscription: user_types,
        content_manager: ContentManager,
        filesystem_manager: FilesystemManager,
        db_manager: SqliteDatabase | None = None,
    ) -> None:
        self.subscription = subscription
        self.content_manager = content_manager
        self.filesystem_manager = filesystem_manager
        self.db_manager = db_manager
        self.metadatas = self.find_metadatas()
        self.update_metadata_filepaths()
        self.redundant_metadatas: list[Path] = []
        self.content_metadatas: list[ContentMetadata] = []
        self.legacy_content_metadatas: list[ContentMetadata] = []

    def merge_content_and_directories(self, unformatted_set: list[dict[str, Any]]):
        temp_content: set[ContentMetadata] = set()
        temp_directories: set[Path] = set()
        for item in unformatted_set:
            if item:
                temp_content.update(item["content"])
                temp_directories.update(item["directories"])
        final_content = list(temp_content)
        final_directories = list(temp_directories)
        final_content.sort(key=lambda obj: obj.content_id)
        return final_content, final_directories

    def find_metadatas(self):
        found_metadatas: list[Path] = []
        for file_ in self.filesystem_manager.get_file_manager(
            self.subscription.id
        ).files:
            if file_.suffix in [".db", ".json"]:
                found_metadatas.append(file_)
        return found_metadatas

    async def process_legacy_metadata(self):
        # Ignore every other metadata fixer, THIS is the one that will merge all legacy metadata (.json and .db) into user_data.
        self.fix_json()
        # Needs work
        # self.fix_sqlite()
        performer_file_manager = self.filesystem_manager.get_file_manager(
            self.subscription.id
        )
        performer_directory_manager = performer_file_manager.directory_manager
        db_manager = DatabaseManager().get_sqlite_db(
            performer_directory_manager.user.metadata_directory.joinpath("user_data.db")
        )
        db_manager.import_metadata(self.legacy_content_metadatas)
        for legacy_filepath in self.metadatas:
            if (
                legacy_filepath.suffix != ".json"
                or "__legacy__" in legacy_filepath.parts
            ):
                continue
            if legacy_filepath and legacy_filepath.exists():
                new_metadata_filepath = (
                    performer_directory_manager.user.metadata_directory.joinpath(
                        "__legacy__", legacy_filepath.name
                    )
                )
                new_metadata_filepath.parent.mkdir(parents=True, exist_ok=True)
                self.filesystem_manager.move(legacy_filepath, new_metadata_filepath)
        await self.filesystem_manager.format_directories(self.subscription)
        await self.fix_archived_db()
        self.metadatas = self.find_metadatas()

    def fix_json(self):
        def merge_statuses(unmerged_status: dict[str, Any]):
            merged_status: list[dict[str, Any]] = [
                val for lst in unmerged_status.values() for val in lst
            ]
            return merged_status

        content_manager = self.content_manager
        for metadata_filepath in self.metadatas:
            if (
                metadata_filepath.suffix != ".json"
                or "__legacy__" in metadata_filepath.parts
                or "Mass Messages" in metadata_filepath.as_posix()
                or "Chats.json" == metadata_filepath.name
            ):
                continue
            final_content_type = None
            archive = False
            new_metadata_set: list[dict[str, Any]] | dict[
                str, Any
            ] = main_helper.import_json(metadata_filepath)
            content_types = self.subscription.get_api().CategorizedContent().get_keys()
            final_stem = metadata_filepath.stem
            patterns = []
            underscore_pattern = r"(_\d+)"
            parentheses_pattern = r"(\(\d+\))"
            patterns.extend([underscore_pattern, parentheses_pattern])

            for pattern in patterns:
                matches = re.findall(pattern, final_stem)
                if matches:
                    final_stem = (
                        final_stem.removesuffix(matches[0]).strip().replace(" ", "")
                    )
            if final_stem in content_types:
                final_content_type = final_stem
            else:
                for item in content_types:
                    if item in metadata_filepath.parts:
                        final_content_type = item
                        break
                if final_stem == "Archived" or "Archived" in metadata_filepath.parts:
                    archive = True
                    final_content_type = "Posts"
                if not final_content_type:
                    # If we get an key error here, we need to move the json file to correct folder or we can resolve it by getting content_type by looking at the directory path
                    # directory_set = set()
                    # merged = merge_statuses(new_metadata_set)
                    # for item in merged:
                    #     directory_set.add(item["directory"])
                    if isinstance(new_metadata_set, dict):
                        if (
                            "type" in new_metadata_set
                            and "content_type" not in new_metadata_set
                        ):
                            item = new_metadata_set["valid"][0]
                            directory = item["directory"]
                            content_types = (
                                self.subscription.get_api().CategorizedContent()
                            )
                            temp_content_type = content_types.path_to_key(
                                Path(directory)
                            )
                            if all(
                                temp_content_type
                                == content_types.path_to_key(Path(item["directory"]))
                                for item in new_metadata_set["valid"]
                            ):
                                final_content_type = temp_content_type
                        else:
                            final_content_type = new_metadata_set["content_type"]
            assert (
                final_content_type
            ), "Content type (Posts,etc) not set before fixing JSON"

            final_metadata_set = {}
            content_json = {}
            if isinstance(new_metadata_set, list):
                for temp_set in new_metadata_set:
                    content_json[temp_set["type"]] = {
                        "valid": temp_set["valid"],
                        "invalid": temp_set["invalid"],
                    }
                final_metadata_set["version"] = 1.8
                final_metadata_set["content"] = content_json
                pass
            elif (
                "type" not in new_metadata_set
                and "version" not in new_metadata_set
                and "valid" in new_metadata_set
            ):
                content_json[metadata_filepath.stem] = {
                    "valid": new_metadata_set["valid"],
                    "invalid": new_metadata_set["invalid"],
                }
                final_metadata_set["version"] = 1.9
                final_metadata_set["content"] = content_json
                pass
            elif "type" in new_metadata_set:
                # Usually means there's another json file we'd have to merge into this, like Images.json and Videos.json.

                class MetadataType(TypedDict):
                    version: float | None
                    content_type: str
                    content: dict[str, Any]

                temp_content_json: MetadataType = {
                    "version": 2.0,
                    "content_type": final_content_type,
                    "content": {},
                }
                temp_content_json["content"][new_metadata_set["type"]] = {
                    "valid": new_metadata_set["valid"],
                    "invalid": new_metadata_set["invalid"],
                }
                content_json = temp_content_json["content"]
                final_metadata_set = temp_content_json
                pass
            else:
                content_json = new_metadata_set.get("content", new_metadata_set)
            if final_metadata_set:
                final_metadata_set["content_type"] = final_content_type
                main_helper.export_json(final_metadata_set, metadata_filepath)
            for media_type, status in content_json.items():
                merged_status = merge_statuses(status)
                if merged_status:
                    for item_json in merged_status:
                        if not final_content_type:
                            raise Exception("content type not found")

                        def assign_metadata(
                            content_metadata: ContentMetadata, item_json: dict[str, Any]
                        ):
                            content_metadata.__legacy__ = True
                            content_metadata.created_at = item_json["postedAt"]
                            content_metadata.paid = item_json.get("paid", False)
                            content_metadata.price = item_json.get("price", 0)
                            content_metadata.text = item_json["text"]
                            content_metadata.user_id = self.subscription.id
                            content_metadata.archived = archive
                            for asset_json in item_json.get("medias", [item_json]):
                                urls = asset_json.get("links") or []
                                if "link" in asset_json:
                                    urls.append(asset_json["link"])

                                media_metadata = content_metadata.find_media(
                                    asset_json.get("media_id"), urls
                                )
                                if not media_metadata:
                                    media_id = asset_json.get("media_id")
                                    if not media_id:
                                        continue
                                    media_metadata = MediaMetadata(
                                        media_id,
                                        media_type,
                                        content_manager,
                                        content_metadata=content_metadata,
                                        urls=urls,
                                    )
                                    pass
                                media_metadata.directory = Path(asset_json["directory"])
                                media_metadata.filename = asset_json["filename"]
                                media_metadata.size = asset_json.get("size", 0)
                                content_metadata.medias.append(media_metadata)

                        content_metadata = None
                        if isinstance(item_json, list):
                            for item_item_json in item_json:
                                content_metadata = self.find_content_metadata_by_id(
                                    item_item_json["post_id"],
                                    content_metadats=self.legacy_content_metadatas,
                                )
                                if not content_metadata:
                                    content_metadata = ContentMetadata(
                                        item_item_json["post_id"],
                                        final_content_type,
                                        content_manager,
                                    )
                                    self.legacy_content_metadatas.append(
                                        content_metadata
                                    )
                                    pass
                                assign_metadata(content_metadata, item_item_json)
                                pass
                            pass
                        else:
                            content_metadata = self.find_content_metadata_by_id(
                                item_json["post_id"],
                                content_metadats=self.legacy_content_metadatas,
                            )
                            if not content_metadata:
                                content_metadata = ContentMetadata(
                                    item_json["post_id"],
                                    final_content_type,
                                    content_manager,
                                )
                                self.legacy_content_metadatas.append(content_metadata)
                            assign_metadata(content_metadata, item_json)

        return self.legacy_content_metadatas

    async def fix_archived_db(
        self,
    ):
        api = self.subscription.get_api()
        filesystem_manager_user = self.filesystem_manager.get_file_manager(
            self.subscription.id
        )
        directory_manager = self.filesystem_manager.get_directory_manager(
            self.subscription.id
        )
        for final_metadata in directory_manager.user.legacy_metadata_directories:
            archived_database_path = final_metadata.joinpath("Archived.db")
            if archived_database_path.exists():
                archived_db_manager = DatabaseManager().get_sqlite_db(
                    archived_database_path, legacy=True
                )
                if not archived_db_manager.session.bind:
                    continue
                for api_type, _value in api.CategorizedContent():
                    database_path = final_metadata.joinpath(f"{api_type}.db")
                    legacy_db_manager = DatabaseManager().get_sqlite_db(
                        database_path, legacy=True
                    )
                    database_name = api_type.lower()
                    result: bool = inspect(
                        archived_db_manager.session.bind.engine
                    ).has_table(database_name)
                    if result:
                        archived_db_manager.alembic_directory = (
                            archived_db_manager.alembic_directory.parent.with_name(
                                database_name
                            ).joinpath("alembic")
                        )
                        archived_db_manager.run_migrations(legacy=True)
                        legacy_db_manager.run_migrations(legacy=True)

                        db_manager = DatabaseManager().get_sqlite_db(database_path)
                        modern_database_session = db_manager.session
                        db_collection = DBCollection()
                        database = db_collection.database_picker("user_data")
                        if not database:
                            return
                        table_name = database.table_picker(api_type, True)
                        if not table_name:
                            return
                        archived_result = archived_db_manager.session.query(
                            table_name
                        ).all()
                        for item in archived_result:
                            result2 = (
                                modern_database_session.query(table_name)
                                .filter(table_name.post_id == item.post_id)  # type: ignore
                                .first()
                            )
                            if not result2:
                                item2 = item.__dict__
                                item2.pop("id")
                                item2.pop("_sa_instance_state")
                                item = table_name(**item2)
                                item.archived = True
                                modern_database_session.add(item)
                            else:
                                result2.archived = True
                        modern_database_session.commit()
                        modern_database_session.close()
                archived_db_manager.session.commit()
                archived_db_manager.session.close()
                new_filepath = Path(
                    archived_database_path.parent,
                    "__legacy__",
                    archived_database_path.name,
                )
                new_filepath.parent.mkdir(exist_ok=True)
                archived_database_path.rename(new_filepath)
                await filesystem_manager_user.remove_file(archived_database_path)
                await filesystem_manager_user.add_file(new_filepath)

    def export(self, api_type: str, datas: list[dict[str, Any]]):
        if api_type == "Posts":
            self.db_manager
        pass

    def find_content_metadata_by_id(
        self, id: int, content_metadats: list[ContentMetadata] = []
    ):
        content_metadats = content_metadats or self.content_metadatas
        for content in content_metadats:
            if content.content_id == id:
                return content

    def update_metadata_filepaths(self):
        filesystem_manager = self.filesystem_manager.get_file_manager(
            self.subscription.id
        )

        directory_manager = filesystem_manager.directory_manager
        if directory_manager:
            for i, metadata_filepath in enumerate(self.metadatas):
                new_m_f = directory_manager.user.metadata_directory.joinpath(
                    metadata_filepath.name
                )
                if metadata_filepath != new_m_f:
                    if (
                        "__legacy__" in metadata_filepath.parts
                        or ".db" != metadata_filepath.suffix
                    ):
                        continue
                    counter = 0
                    while True:
                        if not new_m_f.exists():
                            # If there's metadata present already before the directory is created, we'll create it here
                            directory_manager.user.metadata_directory.mkdir(
                                exist_ok=True, parents=True
                            )
                            self.filesystem_manager.move(metadata_filepath, new_m_f)
                            break
                        else:
                            new_m_f = new_m_f.with_stem(
                                f"{metadata_filepath.stem}_{counter}"
                            )
                            counter += 1
                self.metadatas[i] = new_m_f



# type: ignore

"""content

Revision ID: 0d4d92c0498e
Revises: b791cf213df9
Create Date: 2022-01-14 20:15:27.019051

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "0d4d92c0498e"
down_revision = "b791cf213df9"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("products", sa.Column("title", sa.String(), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("products", "title")
    # ### end Alembic commands ###



# type: ignore
"""content

Revision ID: 29f675c35eee
Revises: 3076beb33c1b
Create Date: 2021-02-04 02:59:01.746229

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "29f675c35eee"
down_revision = "3076beb33c1b"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("medias", schema=None) as batch_op:
        batch_op.add_column(sa.Column("preview", sa.Integer(), nullable=True))

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("medias", schema=None) as batch_op:
        batch_op.drop_column("preview")

    # ### end Alembic commands ###



# type: ignore
"""content

Revision ID: d0118d8ec0b4
Revises: 2c36fcc0b921
Create Date: 2021-02-04 02:59:06.516503

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "d0118d8ec0b4"
down_revision = "2c36fcc0b921"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("medias", schema=None) as batch_op:
        batch_op.add_column(sa.Column("preview", sa.Integer(), nullable=True))

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("medias", schema=None) as batch_op:
        batch_op.drop_column("preview")

    # ### end Alembic commands ###



import random
from typing import Any

import paramiko
from sqlalchemy import MetaData
from sshtunnel import SSHTunnelForwarder  # type: ignore
from ultima_scraper_db.managers.database_manager import Alembica, DatabaseManager


class Project:
    def __init__(
        self,
        name: str,
    ) -> None:
        self.name = name
        self.db_manager = DatabaseManager()

    def handle_ssh(self, db_info: dict[str, Any]):
        ssh_auth_info = db_info["ssh"]
        if ssh_auth_info["host"]:
            private_key_filepath = ssh_auth_info["private_key_filepath"]
            ssh_private_key_password = ssh_auth_info["private_key_password"]
            private_key = (
                paramiko.RSAKey.from_private_key_file(
                    private_key_filepath, ssh_private_key_password
                ).key
                if private_key_filepath
                else None
            )
            random_port = random.randint(6000, 6999)
            ssh_obj = SSHTunnelForwarder(
                (ssh_auth_info["host"], ssh_auth_info["port"]),
                ssh_username=ssh_auth_info["username"],
                ssh_pkey=private_key,
                ssh_private_key_password=ssh_private_key_password,
                remote_bind_address=(db_info["host"], db_info["port"]),
                local_bind_address=(db_info["host"], random_port),
            )
            db_info["ssh"] = ssh_obj
        else:
            db_info["ssh"] = None
        return db_info

    async def _init_db(
        self,
        db_info: dict[str, Any],
        alembica: Alembica,
        metadata: MetaData = MetaData(),
        echo: bool = False,
    ):
        db_info = self.handle_ssh(db_info)
        temp_database = self.db_manager.create_database(
            **db_info, metadata=metadata, alembica=alembica
        )
        self.db_manager.add_database(temp_database)
        await temp_database.init_db(echo)
        return temp_database



from itertools import chain
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from ultima_scraper_api import auth_types
    from ultima_scraper_collection.managers.metadata_manager.metadata_manager import (
        MediaMetadata,
    )


class DefaultCategorizedContent:
    def __init__(self) -> None:
        self.MassMessages: dict[int, dict[str, Any]] = {}
        self.Stories: dict[int, dict[str, Any]] = {}
        self.Posts: dict[int, dict[str, Any]] = {}
        self.Chats: dict[int, dict[str, Any]] = {}
        self.Messages: dict[int, dict[str, Any]] = {}
        self.Highlights: dict[int, dict[str, Any]] = {}

    def __iter__(self):
        for attr, value in self.__dict__.items():
            yield attr, value

    def find_content(self, content_id: int, content_type: str):
        return getattr(self, content_type)[content_id]


class ContentManager:
    def __init__(self, authed: "auth_types") -> None:
        self.authed = authed
        self.auth_session = authed.auth_session
        self.categorized = DefaultCategorizedContent()
        self.media_manager: MediaManager = MediaManager()

    def get_contents(self, content_type: str):
        return getattr(self.categorized, content_type)

    def set_content(self, content_type: str, scraped: list[Any]):
        for content in scraped:
            content_item = getattr(self.categorized, content_type)
            content_item[content.content_id] = content

    def find_content(self, content_id: int, content_type: str):
        found_content = None
        try:
            found_content = getattr(self.categorized, content_type)[content_id]
        except KeyError:
            pass
        return found_content

    def find_media(self, category: str, media_id: int):
        content_items = getattr(self.categorized, category)
        for content in content_items.values():
            for media in content.medias:
                if media.id == media_id:
                    return media

    def get_all_media_ids(self):
        return list(chain(*[x for x in self.categorized.__dict__.values()]))


class MediaManager:
    def __init__(self) -> None:
        self.medias: dict[int, "MediaMetadata"] = {}
        self.invalid_medias: list["MediaMetadata"] = []



import asyncio
from datetime import datetime, timezone
from pathlib import Path
from typing import TYPE_CHECKING, Any

from sqlalchemy import select
from sqlalchemy.orm import joinedload
from ultima_scraper_api.apis.onlyfans.classes.mass_message_model import MassMessageModel
from ultima_scraper_api.apis.onlyfans.onlyfans import OnlyFansAPI
from ultima_scraper_db.databases.ultima_archive.schemas.templates.site import PostModel
from ultima_scraper_renamer.reformat import ReformatManager

from ultima_scraper_collection.config import Sites
from ultima_scraper_collection.managers.metadata_manager.metadata_manager import (
    ApiExtractor,
    ContentMetadata,
)
from ultima_scraper_collection.managers.option_manager import OptionManager
from ultima_scraper_collection.managers.server_manager import ServerManager
from ultima_scraper_collection.modules.module_streamliner import StreamlinedDatascraper

if TYPE_CHECKING:
    from ultima_scraper_api.apis.onlyfans.classes.auth_model import OnlyFansAuthModel
    from ultima_scraper_api.apis.onlyfans.classes.hightlight_model import (
        create_highlight,
    )
    from ultima_scraper_api.apis.onlyfans.classes.message_model import create_message
    from ultima_scraper_api.apis.onlyfans.classes.post_model import create_post
    from ultima_scraper_api.apis.onlyfans.classes.story_model import create_story
    from ultima_scraper_api.apis.onlyfans.classes.user_model import create_user


class OnlyFansDataScraper(StreamlinedDatascraper):
    def __init__(
        self,
        api: OnlyFansAPI,
        option_manager: OptionManager,
        server_manager: ServerManager,
        site_config: Sites.OnlyFansAPIConfig,
    ) -> None:
        self.api = api
        self.option_manager = option_manager
        self.site_config = site_config
        StreamlinedDatascraper.__init__(self, self, server_manager)

    # Scrapes the API for content
    async def media_scraper(
        self,
        content_result: "create_story | create_post | create_message|MassMessageModel",
        subscription: "create_user",
        api_type: str,
    ) -> dict[str, Any]:
        api_type = self.api.convert_api_type_to_key(content_result)
        authed = subscription.get_authed()
        site_config = self.site_config
        new_set: dict[str, Any] = {"content": []}
        directories: list[Path] = []
        if api_type == "Stories":
            pass
        if api_type == "Posts":
            pass
        if api_type == "Messages":
            pass

        content_metadata = ContentMetadata(
            content_result.id, api_type, self.resolve_content_manager(subscription)
        )

        await content_metadata.resolve_extractor(ApiExtractor(content_result))
        for asset in content_metadata.medias:
            if asset.urls:
                reformat_manager = ReformatManager(authed, self.filesystem_manager)
                reformat_item = reformat_manager.prepare_reformat(asset)
                if reformat_item.api_type == "Messages":
                    if (
                        content_metadata.queue_id
                        and content_metadata.__soft__.is_mass_message()
                    ):
                        reformat_item.api_type = "MassMessages"
                file_directory = reformat_item.reformat(
                    site_config.download_setup.directory_format
                )
                reformat_item.directory = file_directory
                file_path = reformat_item.reformat(
                    site_config.download_setup.filename_format
                )
                asset.directory = file_directory
                asset.filename = file_path.name

                if file_directory not in directories:
                    directories.append(file_directory)
        new_set["content"].append(content_metadata)
        new_set["directories"] = directories
        return new_set

    async def get_all_stories(self, subscription: "create_user"):
        """
        get_all_stories(subscription: create_user)

        This function returns a list of all stories and highlights from the given subscription.

        Arguments:
        subscription (create_user): An instance of the create_user class.

        Returns:
        list[create_highlight | create_story]: A list containing all stories and highlights from the subscription.
        """
        master_set: list[create_highlight | create_story] = []
        master_set.extend(await subscription.get_stories())
        master_set.extend(await subscription.get_archived_stories())
        highlights = await subscription.get_highlights()
        valid_highlights: list[create_highlight | create_story] = []
        for highlight in highlights:
            resolved_highlight = await subscription.get_highlights(
                hightlight_id=highlight.id
            )
            valid_highlights.extend(resolved_highlight)
        master_set.extend(valid_highlights)
        return master_set

    async def get_all_posts(self, performer: "create_user") -> list["create_post"]:
        async with self.get_archive_db_api().create_site_api(
            performer.get_api().site_name
        ) as db_site_api:
            after_date = None
            # db_performer = await db_site_api.get_user(performer.id)
            # await db_performer.awaitable_attrs._posts
            # result = await db_performer.last_subscription_downloaded_at()
            # if result:
            #     after_date = result.downloaded_at

            posts = await performer.get_posts(after_date=after_date)
            archived_posts = await performer.get_posts(label="archived")
            private_archived_posts = await performer.get_posts(label="private_archived")

            session = db_site_api.get_session()
            posts_with_comments = (
                select(PostModel)
                .options(joinedload(PostModel.comments))
                .filter(PostModel.comments.any())
                .where(PostModel.user_id == performer.id)
                .order_by(PostModel.created_at.desc())
            )
            results = await session.scalars(posts_with_comments)
            db_posts = results.unique().all()
            threshold_date = (
                db_posts[0].created_at
                if db_posts
                else datetime.min.replace(tzinfo=timezone.utc)
            )
            tasks = [
                x.get_comments()
                for x in performer.scrape_manager.scraped.Posts.values()
                if x.created_at > threshold_date
            ]

            await asyncio.gather(*tasks)
            return posts + archived_posts + private_archived_posts

    async def get_all_subscriptions(
        self,
        authed: "OnlyFansAuthModel",
        identifiers: list[int | str] = [],
        refresh: bool = True,
    ):
        """
        get_all_subscriptions(authed: AuthModel, identifiers: list[int | str] = [], refresh: bool = True)

        This function returns a list of all subscriptions from the given authenticated user.

        Arguments:
        authed (AuthModel): An instance of the AuthModel class.
        identifiers (list[int | str], optional): A list of identifiers (username or id) for the subscriptions. Defaults to an empty list.
        refresh (bool, optional): A flag indicating whether to refresh the list of subscriptions. Defaults to True.

        Returns:
        list[create_subscription]: A list of all subscriptions, sorted by expiredAt, from the authenticated user.
        """
        subscriptions = await authed.get_subscriptions(
            identifiers=identifiers, refresh=refresh, sub_type="active"
        )
        subscriptions.sort(key=lambda x: x.subscribed_by_expire_date)
        return subscriptions



"""content

Revision ID: 37c4f2719d65
Revises: 0d4d92c0498e
Create Date: 2022-03-10 16:00:20.390009

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "37c4f2719d65"
down_revision = "0d4d92c0498e"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "profiles",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("user_id", sa.Integer(), nullable=False),
        sa.Column("username", sa.String(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("username"),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("profiles")
    # ### end Alembic commands ###



# type: ignore
"""content

Revision ID: bf20242a238f
Revises: 7c1c6e101059
Create Date: 2021-06-20 12:42:35.578665

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "bf20242a238f"
down_revision = "7c1c6e101059"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("medias", schema=None) as batch_op:
        batch_op.add_column(sa.Column("api_type", sa.String(), nullable=True))

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("medias", schema=None) as batch_op:
        batch_op.drop_column("api_type")

    # ### end Alembic commands ###



import asyncio
import copy
from pathlib import Path
from urllib.parse import urlparse

import ffmpeg
from aiohttp import ClientResponse
from alive_progress import alive_bar
from ultima_scraper_api import auth_types
from ultima_scraper_api.apis.onlyfans.classes.mass_message_model import MassMessageModel
from ultima_scraper_api.helpers import main_helper
from ultima_scraper_collection.managers.database_manager.connections.sqlite.models.media_model import (
    TemplateMediaModel,
)
from ultima_scraper_collection.managers.filesystem_manager import FilesystemManager
from ultima_scraper_collection.managers.metadata_manager.metadata_manager import (
    MediaMetadata,
)
from ultima_scraper_db.databases.ultima_archive.schemas.templates.site import (
    MediaModel,
    MessageModel,
)
from ultima_scraper_renamer.reformat import ReformatManager


class DownloadManager:
    def __init__(
        self,
        authed: auth_types,
        filesystem_manager: FilesystemManager,
        media_set: set[MediaMetadata] = set(),
        reformat: bool = True,
    ) -> None:
        self.authed = authed
        self.filesystem_manager = filesystem_manager
        self.auth_session = self.authed.auth_session
        self.requester = self.authed.get_requester()
        self.content_list: set[MediaMetadata] = media_set
        self.errors: list[TemplateMediaModel] = []
        self.reformat = reformat
        self.reformat_manager = ReformatManager(self.authed, filesystem_manager)
        self.bar = None

    async def bulk_download(self):
        final_list = [self.download(media_item) for media_item in self.content_list]
        if final_list:
            with alive_bar(len(self.content_list)) as bar:
                self.bar = bar
                _result = await asyncio.gather(*final_list, return_exceptions=True)

    async def drm_download(self, download_item: MediaMetadata):
        content_metadata = download_item.__content_metadata__
        authed = self.authed
        reformat_manager = ReformatManager(authed, self.filesystem_manager)
        assert reformat_manager.filesystem_manager.directory_manager
        site_config = reformat_manager.filesystem_manager.directory_manager.site_config
        drm = authed.drm
        media_item = download_item.__raw__
        assert drm and media_item
        mpd = await drm.get_mpd(media_item)
        pssh = await drm.get_pssh(mpd)
        responses: list[ClientResponse] = []

        if pssh:
            if content_metadata:
                soft_data = content_metadata.__soft__
                raw_data = soft_data.__raw__.copy()
                if (
                    isinstance(soft_data, MassMessageModel)
                    and soft_data
                    and soft_data.author.is_authed_user()
                ):
                    raw_data["responseType"] = ""
            else:
                raw_data = {"responseType": ""}
            license = await drm.get_license(raw_data, media_item, pssh)
            keys = await drm.get_keys(license)
            content_key = keys[-1]
            key = f"{content_key.kid.hex}:{content_key.key.hex()}"
            download_item.key = key
            video_url, audio_url = [
                drm.get_video_url(mpd, media_item),
                drm.get_audio_url(mpd, media_item),
            ]
            download_item.urls = [video_url]
            reformat_item = reformat_manager.prepare_reformat(download_item)
            file_directory = reformat_item.reformat(
                site_config.download_setup.directory_format
            )
            reformat_item.directory = file_directory
            file_path = reformat_item.reformat(
                site_config.download_setup.filename_format
            )
            download_item.directory = file_directory
            download_item.filename = file_path.name
            for media_url in video_url, audio_url:
                drm_download_item = copy.copy(download_item)
                drm_download_item = reformat_manager.drm_format(
                    media_url, drm_download_item
                )

                signature_str = await drm.get_signature(media_item)
                response = await authed.auth_session.request(
                    media_url, premade_settings="", custom_cookies=signature_str
                )
                responses.append(response)
        return responses

    async def download(self, download_item: MediaMetadata):
        if not download_item.urls:
            return
        attempt = 0
        db_media = download_item.__db_media__
        assert db_media
        await db_media.awaitable_attrs.content_media_assos
        content = download_item.get_content_metadata()
        if content:
            db_content = content.__db_content__
            assert db_content
            if isinstance(db_content, MessageModel):
                if db_content.queue_id:
                    try:
                        db_filepath = db_media.find_filepath(
                            (db_content.queue_id, "MassMessages")
                        )
                    except Exception as _e:
                        pass
                    pass
                else:
                    db_filepath = db_media.find_filepath()
            else:
                db_filepath = db_media.find_filepath()
        else:
            db_filepath = db_media.find_filepath()
            pass
        matches = ["us", "uk", "ca", "ca2", "de"]
        p_url = urlparse(download_item.urls[0])
        assert p_url.hostname
        subdomain = p_url.hostname.split(".")[0]
        if any(subdomain in nm for nm in matches):
            return

        authed = self.authed
        authed_drm = authed.drm

        async with self.auth_session.semaphore:
            while attempt < self.auth_session.get_session_manager().max_attempts + 1:
                try:
                    if download_item.drm:
                        if not authed_drm:
                            break
                        responses = await self.drm_download(download_item)
                    else:
                        responses = [
                            await self.requester.request(download_item.urls[0])
                        ]
                    if all(response.status != 200 for response in responses):
                        attempt += 1
                        continue
                    if not download_item.directory:
                        raise Exception(
                            f"{download_item.id} has no directory\n {download_item}"
                        )
                    decrypted_media_paths: list[Path] = []
                    final_size = 0
                    error = None
                    for response in responses:
                        if download_item.drm and await self.drm_check_downloaded(
                            download_item
                        ):
                            continue
                        download_path, error = await self.writer(
                            response, download_item, encrypted=bool(download_item.key)
                        )
                        if error:
                            attempt += 1
                            break
                        if authed_drm and download_item.drm and download_path:
                            output_filepath = authed_drm.decrypt_file(
                                download_path, download_item.key
                            )
                            if not output_filepath:
                                raise Exception("No output_filepath")
                            decrypted_media_paths.append(output_filepath)
                        if response.content_length:
                            final_size += response.content_length
                    if error == 1:
                        # Server Disconnect Error
                        continue
                    elif error == 2:
                        # Resource Not Found Error
                        break
                    assert download_item.filename
                    download_path = download_item.directory.joinpath(
                        download_item.filename
                    )
                    if authed_drm and download_item.drm:
                        formatted = self.format_media(
                            download_path,
                            decrypted_media_paths,
                        )
                        if not formatted:
                            pass
                        final_size = download_path.stat().st_size
                    timestamp = db_media.created_at.timestamp()
                    await main_helper.format_file(
                        download_path, timestamp, self.reformat
                    )
                    if db_media and db_filepath:
                        if not db_filepath.preview:
                            db_media.size = download_item.size = final_size
                        else:
                            if final_size > db_media.size:
                                db_media.size = final_size
                        db_filepath.downloaded = True
                    break
                except asyncio.TimeoutError as _e:
                    continue
                except Exception as _e:
                    print(_e)
        self.bar()

    async def writer(
        self,
        result: ClientResponse,
        download_item: MediaMetadata,
        encrypted: bool = True,
    ):
        async with result as response:
            if download_item.drm and encrypted:
                download_item = copy.copy(download_item)
                download_item = self.reformat_manager.drm_format(
                    response.url.human_repr(), download_item
                )
            assert download_item.directory and download_item.filename
            download_path = Path(download_item.directory, download_item.filename)
            db_media = copy.copy(download_item.__db_media__)
            db_media.directory = download_item.directory
            db_media.filename = download_item.filename
            download = await self.check(db_media, response)
            if not download:
                return download_path, None
            failed = await self.filesystem_manager.write_data(response, download_path)
            return download_path, failed

    async def drm_check_downloaded(self, download_item: MediaMetadata):
        download_path = download_item.get_filepath()
        if download_path.exists():
            if download_path.stat().st_size and download_item.__db_media__.size:
                return True
        return False

    async def check(self, download_item: MediaModel, response: ClientResponse):
        # Checks if we should download item or not // True | False
        filepath = Path(download_item.directory, download_item.filename)
        response_status = False
        if response.status == 200:
            response_status = True
            if response.content_length:
                download_item.size = response.content_length

        if filepath.exists():
            try:
                if filepath.stat().st_size == response.content_length:
                    return False
                else:
                    return True
            except Exception as _e:
                pass
        else:
            if response_status:
                # Can produce false positives due to the same reason below
                return True
            else:
                # Reached this point because it probably exists in the folder but under a different content category
                pass

    def format_media(self, output_filepath: Path, decrypted_media_paths: list[Path]):
        # If you have decrypted video and audio to merge
        if len(decrypted_media_paths) > 1:
            dec_video_path, dec_audio_path = decrypted_media_paths
            video_input = ffmpeg.input(dec_video_path)  # type:ignore
            audio_input = ffmpeg.input(dec_audio_path)  # type:ignore
            try:
                _ffmpeg_output = ffmpeg.output(  # type:ignore
                    video_input,  # type:ignore
                    audio_input,  # type:ignore
                    output_filepath.as_posix(),
                    vcodec="copy",
                    acodec="copy",
                ).run(capture_stdout=True, capture_stderr=True, overwrite_output=True)
                return True
            except ffmpeg.Error as _e:
                return False
        return True



from pathlib import Path

from ultima_scraper_collection.managers.database_manager.connections.sqlite.sqlite_database import (
    SqliteDatabase,
)


class DatabaseManager:
    def __init__(self) -> None:
        self.active_db: SqliteDatabase | None = None

    def get_sqlite_db(self, path: Path, legacy: bool = False):
        sqlite_db = SqliteDatabase().init_db(path, legacy)
        self.active_db = sqlite_db
        return sqlite_db



import ultima_scraper_api
from ultima_scraper_api import SUPPORTED_SITES
from ultima_scraper_api.apis.onlyfans import onlyfans
from ultima_scraper_collection import datascraper_types
from ultima_scraper_collection.config import UltimaScraperCollectionConfig
from ultima_scraper_collection.managers.datascraper_manager.datascrapers.fansly import (
    FanslyDataScraper,
)
from ultima_scraper_collection.managers.datascraper_manager.datascrapers.onlyfans import (
    OnlyFansDataScraper,
)
from ultima_scraper_collection.managers.option_manager import OptionManager
from ultima_scraper_collection.managers.server_manager import ServerManager


class DataScraperManager:
    def __init__(
        self, server_manager: ServerManager, config: UltimaScraperCollectionConfig
    ) -> None:
        self.datascrapers: dict[str, datascraper_types] = {}
        self.server_manager: ServerManager = server_manager
        self.config = config
        for site_name in SUPPORTED_SITES:
            datascraper = self.add_datascraper(
                ultima_scraper_api.select_api(site_name, config),
                OptionManager(),
                self.server_manager,
            )
            datascraper.filesystem_manager.activate_directory_manager(
                self.get_site_config(site_name)
            )

    def get_site_config(self, name: str):
        return getattr(self.config.site_apis, name.lower())

    def find_datascraper(
        self,
        site_name: str,
    ):
        return self.datascrapers.get(site_name.lower())

    def select_datascraper(
        self,
        site_name: str,
    ):
        return self.datascrapers.get(site_name.lower())

    def add_datascraper(
        self,
        api: ultima_scraper_api.api_types,
        option_manager: OptionManager,
        server_manager: ServerManager,
    ):
        site_settings = self.get_site_config(api.site_name)
        if isinstance(api, onlyfans.OnlyFansAPI):
            datascraper = OnlyFansDataScraper(
                api, option_manager, server_manager, site_settings
            )
        else:
            datascraper = FanslyDataScraper(
                api, option_manager, server_manager, site_settings
            )
        self.datascrapers[api.site_name.lower()] = datascraper
        return datascraper



# type: ignore
"""content

Revision ID: 7c1c6e101059
Revises: aeb9fe314556
Create Date: 2021-05-31 02:56:29.998095

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "7c1c6e101059"
down_revision = "aeb9fe314556"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("medias", schema=None) as batch_op:
        batch_op.alter_column(
            "created_at",
            existing_type=sa.DATETIME(),
            type_=sa.TIMESTAMP(),
            existing_nullable=True,
        )

    with op.batch_alter_table("messages", schema=None) as batch_op:
        batch_op.add_column(sa.Column("archived", sa.Boolean(), nullable=True))
        batch_op.alter_column(
            "created_at",
            existing_type=sa.DATETIME(),
            type_=sa.TIMESTAMP(),
            existing_nullable=True,
        )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("messages", schema=None) as batch_op:
        batch_op.alter_column(
            "created_at",
            existing_type=sa.TIMESTAMP(),
            type_=sa.DATETIME(),
            existing_nullable=True,
        )
        batch_op.drop_column("archived")

    with op.batch_alter_table("medias", schema=None) as batch_op:
        batch_op.alter_column(
            "created_at",
            existing_type=sa.TIMESTAMP(),
            type_=sa.DATETIME(),
            existing_nullable=True,
        )

    # ### end Alembic commands ###
